"""
This script was generated from jaaql.fxli at 12/03/2023, 15:27:49
"""

from jaaql.db.db_interface import DBInterface
from jaaql.db.db_utils import execute_supplied_statement, execute_supplied_statement_singleton


# Generated keys for table 'application'
KG__application__name = "name"
KG__application__base_url = "base_url"
KG__application__artifacts_source = "artifacts_source"
KG__application__default_schema = "default_schema"
KG__application__default_s_et = "default_s_et"
KG__application__default_a_et = "default_a_et"
KG__application__default_r_et = "default_r_et"
KG__application__default_u_et = "default_u_et"
KG__application__unlock_key_validity_period = "unlock_key_validity_period"
KG__application__unlock_code_validity_period = "unlock_code_validity_period"
KG__application__is_live = "is_live"

# Generated queries for table 'application'
QG__application_delete = "DELETE FROM application WHERE name = :name"
QG__application_insert = """
    INSERT INTO application (name, base_url, artifacts_source,
        default_schema, default_s_et, default_a_et,
        default_r_et, default_u_et)
    VALUES (:name, :base_url, :artifacts_source,
        :default_schema, :default_s_et, :default_a_et,
        :default_r_et, :default_u_et)
    RETURNING unlock_key_validity_period, unlock_code_validity_period, is_live
"""
QG__application_select_all = "SELECT * FROM application"
QG__application_select = "SELECT * FROM application WHERE name = :name"
QG__application_update = """
    UPDATE
        application
    SET
        base_url = COALESCE(:base_url, base_url),
        artifacts_source = COALESCE(:artifacts_source, artifacts_source),
        default_schema = COALESCE(:default_schema, default_schema),
        default_s_et = COALESCE(:default_s_et, default_s_et),
        default_a_et = COALESCE(:default_a_et, default_a_et),
        default_r_et = COALESCE(:default_r_et, default_r_et),
        default_u_et = COALESCE(:default_u_et, default_u_et),
        unlock_key_validity_period = COALESCE(:unlock_key_validity_period, unlock_key_validity_period),
        unlock_code_validity_period = COALESCE(:unlock_code_validity_period, unlock_code_validity_period),
        is_live = COALESCE(:is_live, is_live)
    WHERE
        name = :name
"""


# Generated methods for table 'application'
def application__delete(
    connection: DBInterface,
    name
):
    execute_supplied_statement(
        connection, QG__application_delete,
        {
            # Key Fields
            KG__application__name: name
        }
    )


def application__update(
    connection: DBInterface,
    name,
    base_url=None, artifacts_source=None, default_schema=None,
    default_s_et=None, default_a_et=None, default_r_et=None,
    default_u_et=None, unlock_key_validity_period=None, unlock_code_validity_period=None,
    is_live=None
):
    execute_supplied_statement(
        connection, QG__application_update,
        {
            # Key Fields
            KG__application__name: name,

            # None Key Fields
            KG__application__base_url: base_url,
            KG__application__artifacts_source: artifacts_source,
            KG__application__default_schema: default_schema,
            KG__application__default_s_et: default_s_et,
            KG__application__default_a_et: default_a_et,
            KG__application__default_r_et: default_r_et,
            KG__application__default_u_et: default_u_et,
            KG__application__unlock_key_validity_period: unlock_key_validity_period,
            KG__application__unlock_code_validity_period: unlock_code_validity_period,
            KG__application__is_live: is_live
        }
    )


def application__select(
    connection: DBInterface,
    name,
    singleton_code: int = None, singleton_message: str = "application does not exist"
):
    return execute_supplied_statement_singleton(
        connection, QG__application_select, 
        {
            # Key Fields
            KG__application__name: name
        },
        as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


def application__select_all(
    connection: DBInterface
):
    return execute_supplied_statement(
        connection, QG__application_select_all,
        as_objects=True
    )


def application__insert(
    connection: DBInterface,
    name, base_url,
    artifacts_source=None, default_schema=None, default_s_et=None,
    default_a_et=None, default_r_et=None, default_u_et=None
):
    return execute_supplied_statement_singleton(
        connection, QG__application_insert,
        {
            KG__application__name: name,
            KG__application__base_url: base_url,
            KG__application__artifacts_source: artifacts_source,
            KG__application__default_schema: default_schema,
            KG__application__default_s_et: default_s_et,
            KG__application__default_a_et: default_a_et,
            KG__application__default_r_et: default_r_et,
            KG__application__default_u_et: default_u_et
        }, as_objects=True
    )


# Generated keys for table 'application_schema'
KG__application_schema__application = "application"
KG__application_schema__name = "name"
KG__application_schema__database = "database"

# Generated queries for table 'application_schema'
QG__application_schema_delete = "DELETE FROM application_schema WHERE application = :application AND name = :name"
QG__application_schema_insert = """
    INSERT INTO application_schema (application, name, database)
    VALUES (:application, :name, :database)
"""
QG__application_schema_select_all = "SELECT * FROM application_schema"
QG__application_schema_select = "SELECT * FROM application_schema WHERE application = :application AND name = :name"
QG__application_schema_update = """
    UPDATE
        application_schema
    SET
        database = COALESCE(:database, database)
    WHERE
        application = :application AND name = :name
"""


# Generated methods for table 'application_schema'
def application_schema__delete(
    connection: DBInterface,
    application, name
):
    execute_supplied_statement(
        connection, QG__application_schema_delete,
        {
            # Key Fields
            KG__application_schema__application: application,
            KG__application_schema__name: name
        }
    )


def application_schema__update(
    connection: DBInterface,
    application, name,
    database=None
):
    execute_supplied_statement(
        connection, QG__application_schema_update,
        {
            # Key Fields
            KG__application_schema__application: application,
            KG__application_schema__name: name,

            # None Key Fields
            KG__application_schema__database: database
        }
    )


def application_schema__select(
    connection: DBInterface,
    application, name,
    singleton_code: int = None, singleton_message: str = "application_schema does not exist"
):
    return execute_supplied_statement_singleton(
        connection, QG__application_schema_select, 
        {
            # Key Fields
            KG__application_schema__application: application,
            KG__application_schema__name: name
        },
        as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


def application_schema__select_all(
    connection: DBInterface
):
    return execute_supplied_statement(
        connection, QG__application_schema_select_all,
        as_objects=True
    )


def application_schema__insert(
    connection: DBInterface,
    application, name, database
):
    execute_supplied_statement(
        connection, QG__application_schema_insert,
        {
            KG__application_schema__application: application,
            KG__application_schema__name: name,
            KG__application_schema__database: database
        }
    )


# Generated keys for table 'email_dispatcher'
KG__email_dispatcher__application = "application"
KG__email_dispatcher__name = "name"
KG__email_dispatcher__display_name = "display_name"
KG__email_dispatcher__protocol = "protocol"
KG__email_dispatcher__url = "url"
KG__email_dispatcher__port = "port"
KG__email_dispatcher__username = "username"
KG__email_dispatcher__password = "password"
KG__email_dispatcher__whitelist = "whitelist"

# Generated queries for table 'email_dispatcher'
QG__email_dispatcher_delete = "DELETE FROM email_dispatcher WHERE application = :application AND name = :name"
QG__email_dispatcher_insert = """
    INSERT INTO email_dispatcher (application, name, display_name,
        protocol, url, port,
        username, password, whitelist)
    VALUES (:application, :name, :display_name,
        :protocol, :url, :port,
        :username, :password, :whitelist)
"""
QG__email_dispatcher_select_all = "SELECT * FROM email_dispatcher"
QG__email_dispatcher_select = "SELECT * FROM email_dispatcher WHERE application = :application AND name = :name"
QG__email_dispatcher_update = """
    UPDATE
        email_dispatcher
    SET
        display_name = COALESCE(:display_name, display_name),
        protocol = COALESCE(:protocol, protocol),
        url = COALESCE(:url, url),
        port = COALESCE(:port, port),
        username = COALESCE(:username, username),
        password = COALESCE(:password, password),
        whitelist = COALESCE(:whitelist, whitelist)
    WHERE
        application = :application AND name = :name
"""


# Generated methods for table 'email_dispatcher'
def email_dispatcher__delete(
    connection: DBInterface,
    application, name
):
    execute_supplied_statement(
        connection, QG__email_dispatcher_delete,
        {
            # Key Fields
            KG__email_dispatcher__application: application,
            KG__email_dispatcher__name: name
        }
    )


def email_dispatcher__update(
    connection: DBInterface, encryption_key: bytes,
    application, name,
    display_name=None, protocol=None, url=None,
    port=None, username=None, password=None,
    whitelist=None,
    encryption_salts=None
):
    execute_supplied_statement(
        connection, QG__email_dispatcher_update,
        {
            # Key Fields
            KG__email_dispatcher__application: application,
            KG__email_dispatcher__name: name,

            # None Key Fields
            KG__email_dispatcher__display_name: display_name,
            KG__email_dispatcher__protocol: protocol,
            KG__email_dispatcher__url: url,
            KG__email_dispatcher__port: port,
            KG__email_dispatcher__username: username,
            KG__email_dispatcher__password: password,
            KG__email_dispatcher__whitelist: whitelist
        }, encryption_key=encryption_key, encryption_salts=encryption_salts, encrypt_parameters=[
            KG__email_dispatcher__password
        ]
    )


def email_dispatcher__select(
    connection: DBInterface, encryption_key: bytes,
    application, name,
    singleton_code: int = None, singleton_message: str = "email_dispatcher does not exist"
):
    return execute_supplied_statement_singleton(
        connection, QG__email_dispatcher_select, 
        {
            # Key Fields
            KG__email_dispatcher__application: application,
            KG__email_dispatcher__name: name
        }, encryption_key=encryption_key, decrypt_columns=[
            KG__email_dispatcher__password
        ],
        as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


def email_dispatcher__select_all(
    connection: DBInterface, encryption_key: bytes
):
    return execute_supplied_statement(
        connection, QG__email_dispatcher_select_all, encryption_key=encryption_key, decrypt_columns=[
            KG__email_dispatcher__password
        ],
        as_objects=True
    )


def email_dispatcher__insert(
    connection: DBInterface, encryption_key: bytes,
    application, name, display_name,
    protocol=None, url=None, port=None,
    username=None, password=None, whitelist=None,
    encryption_salts=None
):
    execute_supplied_statement(
        connection, QG__email_dispatcher_insert,
        {
            KG__email_dispatcher__application: application,
            KG__email_dispatcher__name: name,
            KG__email_dispatcher__display_name: display_name,
            KG__email_dispatcher__protocol: protocol,
            KG__email_dispatcher__url: url,
            KG__email_dispatcher__port: port,
            KG__email_dispatcher__username: username,
            KG__email_dispatcher__password: password,
            KG__email_dispatcher__whitelist: whitelist
        }, encryption_key=encryption_key, encryption_salts=encryption_salts, encrypt_parameters=[
            KG__email_dispatcher__password
        ]
    )


# Generated keys for table 'jaaql'
KG__jaaql__the_anonymous_user = "the_anonymous_user"
KG__jaaql__security_event_attempt_limit = "security_event_attempt_limit"

# Generated queries for table 'jaaql'
QG__jaaql_delete = "DELETE FROM jaaql"
QG__jaaql_insert = """
    INSERT INTO jaaql (the_anonymous_user, security_event_attempt_limit)
    VALUES (:the_anonymous_user, :security_event_attempt_limit)
"""
QG__jaaql_select_all = "SELECT * FROM jaaql"
QG__jaaql_select = "SELECT * FROM jaaql"
QG__jaaql_update = """
    UPDATE
        jaaql
    SET
        the_anonymous_user = COALESCE(:the_anonymous_user, the_anonymous_user),
        security_event_attempt_limit = COALESCE(:security_event_attempt_limit, security_event_attempt_limit)
"""


# Generated methods for table 'jaaql'
def jaaql__delete(
    connection: DBInterface
):
    execute_supplied_statement(
        connection, QG__jaaql_delete)


def jaaql__update(
    connection: DBInterface,
    the_anonymous_user=None, security_event_attempt_limit=None
):
    execute_supplied_statement(
        connection, QG__jaaql_update,
        {
            # None Key Fields
            KG__jaaql__the_anonymous_user: the_anonymous_user,
            KG__jaaql__security_event_attempt_limit: security_event_attempt_limit
        }
    )


def jaaql__select(
    connection: DBInterface,
    singleton_code: int = None, singleton_message: str = "jaaql does not exist"
):
    return execute_supplied_statement_singleton(
        connection, QG__jaaql_select,
        as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


def jaaql__select_all(
    connection: DBInterface
):
    return execute_supplied_statement(
        connection, QG__jaaql_select_all,
        as_objects=True
    )


def jaaql__insert(
    connection: DBInterface,
    the_anonymous_user, security_event_attempt_limit
):
    execute_supplied_statement(
        connection, QG__jaaql_insert,
        {
            KG__jaaql__the_anonymous_user: the_anonymous_user,
            KG__jaaql__security_event_attempt_limit: security_event_attempt_limit
        }
    )


# Generated keys for table 'email_template'
KG__email_template__application = "application"
KG__email_template__dispatcher = "dispatcher"
KG__email_template__name = "name"
KG__email_template__type = "type"
KG__email_template__content_url = "content_url"
KG__email_template__validation_schema = "validation_schema"
KG__email_template__data_validation_table = "data_validation_table"
KG__email_template__data_validation_view = "data_validation_view"

# Generated queries for table 'email_template'
QG__email_template_delete = "DELETE FROM email_template WHERE application = :application AND name = :name"
QG__email_template_insert = """
    INSERT INTO email_template (application, dispatcher, name,
        type, content_url, validation_schema,
        data_validation_table, data_validation_view)
    VALUES (:application, :dispatcher, :name,
        :type, :content_url, :validation_schema,
        :data_validation_table, :data_validation_view)
"""
QG__email_template_select_all = "SELECT * FROM email_template"
QG__email_template_select = "SELECT * FROM email_template WHERE application = :application AND name = :name"
QG__email_template_update = """
    UPDATE
        email_template
    SET
        dispatcher = COALESCE(:dispatcher, dispatcher),
        type = COALESCE(:type, type),
        content_url = COALESCE(:content_url, content_url),
        validation_schema = COALESCE(:validation_schema, validation_schema),
        data_validation_table = COALESCE(:data_validation_table, data_validation_table),
        data_validation_view = COALESCE(:data_validation_view, data_validation_view)
    WHERE
        application = :application AND name = :name
"""


# Generated methods for table 'email_template'
def email_template__delete(
    connection: DBInterface,
    application, name
):
    execute_supplied_statement(
        connection, QG__email_template_delete,
        {
            # Key Fields
            KG__email_template__application: application,
            KG__email_template__name: name
        }
    )


def email_template__update(
    connection: DBInterface,
    application, name,
    dispatcher=None, type=None, content_url=None,
    validation_schema=None, data_validation_table=None, data_validation_view=None
):
    execute_supplied_statement(
        connection, QG__email_template_update,
        {
            # Key Fields
            KG__email_template__application: application,
            KG__email_template__name: name,

            # None Key Fields
            KG__email_template__dispatcher: dispatcher,
            KG__email_template__type: type,
            KG__email_template__content_url: content_url,
            KG__email_template__validation_schema: validation_schema,
            KG__email_template__data_validation_table: data_validation_table,
            KG__email_template__data_validation_view: data_validation_view
        }
    )


def email_template__select(
    connection: DBInterface,
    application, name,
    singleton_code: int = None, singleton_message: str = "email_template does not exist"
):
    return execute_supplied_statement_singleton(
        connection, QG__email_template_select, 
        {
            # Key Fields
            KG__email_template__application: application,
            KG__email_template__name: name
        },
        as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


def email_template__select_all(
    connection: DBInterface
):
    return execute_supplied_statement(
        connection, QG__email_template_select_all,
        as_objects=True
    )


def email_template__insert(
    connection: DBInterface,
    application, dispatcher, name,
    type, content_url,
    validation_schema=None, data_validation_table=None, data_validation_view=None
):
    execute_supplied_statement(
        connection, QG__email_template_insert,
        {
            KG__email_template__application: application,
            KG__email_template__dispatcher: dispatcher,
            KG__email_template__name: name,
            KG__email_template__type: type,
            KG__email_template__content_url: content_url,
            KG__email_template__validation_schema: validation_schema,
            KG__email_template__data_validation_table: data_validation_table,
            KG__email_template__data_validation_view: data_validation_view
        }
    )


# Generated keys for table 'document_template'
KG__document_template__application = "application"
KG__document_template__name = "name"
KG__document_template__content_path = "content_path"
KG__document_template__email_template = "email_template"

# Generated queries for table 'document_template'
QG__document_template_delete = "DELETE FROM document_template WHERE application = :application AND name = :name"
QG__document_template_insert = """
    INSERT INTO document_template (application, name, content_path,
        email_template)
    VALUES (:application, :name, :content_path,
        :email_template)
"""
QG__document_template_select_all = "SELECT * FROM document_template"
QG__document_template_select = "SELECT * FROM document_template WHERE application = :application AND name = :name"
QG__document_template_update = """
    UPDATE
        document_template
    SET
        content_path = COALESCE(:content_path, content_path),
        email_template = COALESCE(:email_template, email_template)
    WHERE
        application = :application AND name = :name
"""


# Generated methods for table 'document_template'
def document_template__delete(
    connection: DBInterface,
    application, name
):
    execute_supplied_statement(
        connection, QG__document_template_delete,
        {
            # Key Fields
            KG__document_template__application: application,
            KG__document_template__name: name
        }
    )


def document_template__update(
    connection: DBInterface,
    application, name,
    content_path=None, email_template=None
):
    execute_supplied_statement(
        connection, QG__document_template_update,
        {
            # Key Fields
            KG__document_template__application: application,
            KG__document_template__name: name,

            # None Key Fields
            KG__document_template__content_path: content_path,
            KG__document_template__email_template: email_template
        }
    )


def document_template__select(
    connection: DBInterface,
    application, name,
    singleton_code: int = None, singleton_message: str = "document_template does not exist"
):
    return execute_supplied_statement_singleton(
        connection, QG__document_template_select, 
        {
            # Key Fields
            KG__document_template__application: application,
            KG__document_template__name: name
        },
        as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


def document_template__select_all(
    connection: DBInterface
):
    return execute_supplied_statement(
        connection, QG__document_template_select_all,
        as_objects=True
    )


def document_template__insert(
    connection: DBInterface,
    application, name, content_path,
    email_template=None
):
    execute_supplied_statement(
        connection, QG__document_template_insert,
        {
            KG__document_template__application: application,
            KG__document_template__name: name,
            KG__document_template__content_path: content_path,
            KG__document_template__email_template: email_template
        }
    )


# Generated keys for table 'document_request'
KG__document_request__application = "application"
KG__document_request__template = "template"
KG__document_request__uuid = "uuid"
KG__document_request__request_timestamp = "request_timestamp"
KG__document_request__encrypted_access_token = "encrypted_access_token"
KG__document_request__encrypted_parameters = "encrypted_parameters"
KG__document_request__render_timestamp = "render_timestamp"

# Generated queries for table 'document_request'
QG__document_request_delete = "DELETE FROM document_request WHERE uuid = :uuid"
QG__document_request_insert = """
    INSERT INTO document_request (application, template, uuid,
        encrypted_access_token, encrypted_parameters, render_timestamp)
    VALUES (:application, :template, :uuid,
        :encrypted_access_token, :encrypted_parameters, :render_timestamp)
    RETURNING request_timestamp
"""
QG__document_request_select_all = "SELECT * FROM document_request"
QG__document_request_select = "SELECT * FROM document_request WHERE uuid = :uuid"
QG__document_request_update = """
    UPDATE
        document_request
    SET
        application = COALESCE(:application, application),
        template = COALESCE(:template, template),
        request_timestamp = COALESCE(:request_timestamp, request_timestamp),
        encrypted_access_token = COALESCE(:encrypted_access_token, encrypted_access_token),
        encrypted_parameters = COALESCE(:encrypted_parameters, encrypted_parameters),
        render_timestamp = COALESCE(:render_timestamp, render_timestamp)
    WHERE
        uuid = :uuid
"""


# Generated methods for table 'document_request'
def document_request__delete(
    connection: DBInterface,
    uuid
):
    execute_supplied_statement(
        connection, QG__document_request_delete,
        {
            # Key Fields
            KG__document_request__uuid: uuid
        }
    )


def document_request__update(
    connection: DBInterface, encryption_key: bytes,
    uuid,
    application=None, template=None, request_timestamp=None,
    encrypted_access_token=None, encrypted_parameters=None, render_timestamp=None,
    encryption_salts=None
):
    execute_supplied_statement(
        connection, QG__document_request_update,
        {
            # Key Fields
            KG__document_request__uuid: uuid,

            # None Key Fields
            KG__document_request__application: application,
            KG__document_request__template: template,
            KG__document_request__request_timestamp: request_timestamp,
            KG__document_request__encrypted_access_token: encrypted_access_token,
            KG__document_request__encrypted_parameters: encrypted_parameters,
            KG__document_request__render_timestamp: render_timestamp
        }, encryption_key=encryption_key, encryption_salts=encryption_salts, encrypt_parameters=[
            KG__document_request__encrypted_access_token,
            KG__document_request__encrypted_parameters
        ]
    )


def document_request__select(
    connection: DBInterface, encryption_key: bytes,
    uuid,
    singleton_code: int = None, singleton_message: str = "document_request does not exist"
):
    return execute_supplied_statement_singleton(
        connection, QG__document_request_select, 
        {
            # Key Fields
            KG__document_request__uuid: uuid
        }, encryption_key=encryption_key, decrypt_columns=[
            KG__document_request__encrypted_access_token,
            KG__document_request__encrypted_parameters
        ],
        as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


def document_request__select_all(
    connection: DBInterface, encryption_key: bytes
):
    return execute_supplied_statement(
        connection, QG__document_request_select_all, encryption_key=encryption_key, decrypt_columns=[
            KG__document_request__encrypted_access_token,
            KG__document_request__encrypted_parameters
        ],
        as_objects=True
    )


def document_request__insert(
    connection: DBInterface, encryption_key: bytes,
    application, template, uuid,
    encrypted_access_token,
    encrypted_parameters=None, render_timestamp=None,
    encryption_salts=None
):
    return execute_supplied_statement_singleton(
        connection, QG__document_request_insert,
        {
            KG__document_request__application: application,
            KG__document_request__template: template,
            KG__document_request__uuid: uuid,
            KG__document_request__encrypted_access_token: encrypted_access_token,
            KG__document_request__encrypted_parameters: encrypted_parameters,
            KG__document_request__render_timestamp: render_timestamp
        }, encryption_key=encryption_key, encryption_salts=encryption_salts, encrypt_parameters=[
            KG__document_request__encrypted_access_token,
            KG__document_request__encrypted_parameters
        ], as_objects=True
    )


# Generated keys for table 'account'
KG__account__id = "id"
KG__account__username = "username"
KG__account__deletion_timestamp = "deletion_timestamp"
KG__account__most_recent_password = "most_recent_password"

# Generated queries for table 'account'
QG__account_delete = "DELETE FROM account WHERE id = :id"
QG__account_insert = """
    INSERT INTO account (id, username, deletion_timestamp,
        most_recent_password)
    VALUES (:id, :username, :deletion_timestamp,
        :most_recent_password)
"""
QG__account_select_all = "SELECT * FROM account"
QG__account_select = "SELECT * FROM account WHERE id = :id"
QG__account_update = """
    UPDATE
        account
    SET
        username = COALESCE(:username, username),
        deletion_timestamp = COALESCE(:deletion_timestamp, deletion_timestamp),
        most_recent_password = COALESCE(:most_recent_password, most_recent_password)
    WHERE
        id = :id
"""


# Generated methods for table 'account'
def account__delete(
    connection: DBInterface,
    id
):
    execute_supplied_statement(
        connection, QG__account_delete,
        {
            # Key Fields
            KG__account__id: id
        }
    )


def account__update(
    connection: DBInterface, encryption_key: bytes,
    id,
    username=None, deletion_timestamp=None, most_recent_password=None,
    encryption_salts=None
):
    execute_supplied_statement(
        connection, QG__account_update,
        {
            # Key Fields
            KG__account__id: id,

            # None Key Fields
            KG__account__username: username,
            KG__account__deletion_timestamp: deletion_timestamp,
            KG__account__most_recent_password: most_recent_password
        }, encryption_key=encryption_key, encryption_salts=encryption_salts, encrypt_parameters=[
            KG__account__username
        ]
    )


def account__select(
    connection: DBInterface, encryption_key: bytes,
    id,
    singleton_code: int = None, singleton_message: str = "account does not exist"
):
    return execute_supplied_statement_singleton(
        connection, QG__account_select, 
        {
            # Key Fields
            KG__account__id: id
        }, encryption_key=encryption_key, decrypt_columns=[
            KG__account__username
        ],
        as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


def account__select_all(
    connection: DBInterface, encryption_key: bytes
):
    return execute_supplied_statement(
        connection, QG__account_select_all, encryption_key=encryption_key, decrypt_columns=[
            KG__account__username
        ],
        as_objects=True
    )


def account__insert(
    connection: DBInterface, encryption_key: bytes,
    id, username,
    deletion_timestamp=None, most_recent_password=None,
    encryption_salts=None
):
    execute_supplied_statement(
        connection, QG__account_insert,
        {
            KG__account__id: id,
            KG__account__username: username,
            KG__account__deletion_timestamp: deletion_timestamp,
            KG__account__most_recent_password: most_recent_password
        }, encryption_key=encryption_key, encryption_salts=encryption_salts, encrypt_parameters=[
            KG__account__username
        ]
    )


# Generated keys for table 'account_password'
KG__account_password__account = "account"
KG__account_password__uuid = "uuid"
KG__account_password__hash = "hash"
KG__account_password__creation_timestamp = "creation_timestamp"

# Generated queries for table 'account_password'
QG__account_password_delete = "DELETE FROM account_password WHERE uuid = :uuid"
QG__account_password_insert = """
    INSERT INTO account_password (account, hash)
    VALUES (:account, :hash)
    RETURNING uuid, creation_timestamp
"""
QG__account_password_select_all = "SELECT * FROM account_password"
QG__account_password_select = "SELECT * FROM account_password WHERE uuid = :uuid"
QG__account_password_update = """
    UPDATE
        account_password
    SET
        account = COALESCE(:account, account),
        hash = COALESCE(:hash, hash),
        creation_timestamp = COALESCE(:creation_timestamp, creation_timestamp)
    WHERE
        uuid = :uuid
"""


# Generated methods for table 'account_password'
def account_password__delete(
    connection: DBInterface,
    uuid
):
    execute_supplied_statement(
        connection, QG__account_password_delete,
        {
            # Key Fields
            KG__account_password__uuid: uuid
        }
    )


def account_password__update(
    connection: DBInterface, encryption_key: bytes,
    uuid,
    account=None, hash=None, creation_timestamp=None,
    encryption_salts=None
):
    execute_supplied_statement(
        connection, QG__account_password_update,
        {
            # Key Fields
            KG__account_password__uuid: uuid,

            # None Key Fields
            KG__account_password__account: account,
            KG__account_password__hash: hash,
            KG__account_password__creation_timestamp: creation_timestamp
        }, encryption_key=encryption_key, encryption_salts=encryption_salts, encrypt_parameters=[
            KG__account_password__hash
        ]
    )


def account_password__select(
    connection: DBInterface, encryption_key: bytes,
    uuid,
    singleton_code: int = None, singleton_message: str = "account_password does not exist"
):
    return execute_supplied_statement_singleton(
        connection, QG__account_password_select, 
        {
            # Key Fields
            KG__account_password__uuid: uuid
        }, encryption_key=encryption_key, decrypt_columns=[
            KG__account_password__hash
        ],
        as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


def account_password__select_all(
    connection: DBInterface, encryption_key: bytes
):
    return execute_supplied_statement(
        connection, QG__account_password_select_all, encryption_key=encryption_key, decrypt_columns=[
            KG__account_password__hash
        ],
        as_objects=True
    )


def account_password__insert(
    connection: DBInterface, encryption_key: bytes,
    account, hash,
    encryption_salts=None
):
    return execute_supplied_statement_singleton(
        connection, QG__account_password_insert,
        {
            KG__account_password__account: account,
            KG__account_password__hash: hash
        }, encryption_key=encryption_key, encryption_salts=encryption_salts, encrypt_parameters=[
            KG__account_password__hash
        ], as_objects=True
    )


# Generated keys for table 'validated_ip_address'
KG__validated_ip_address__account = "account"
KG__validated_ip_address__uuid = "uuid"
KG__validated_ip_address__encrypted_salted_ip_address = "encrypted_salted_ip_address"
KG__validated_ip_address__first_authentication_timestamp = "first_authentication_timestamp"
KG__validated_ip_address__last_authentication_timestamp = "last_authentication_timestamp"

# Generated queries for table 'validated_ip_address'
QG__validated_ip_address_delete = "DELETE FROM validated_ip_address WHERE uuid = :uuid"
QG__validated_ip_address_insert = """
    INSERT INTO validated_ip_address (account, encrypted_salted_ip_address, last_authentication_timestamp)
    VALUES (:account, :encrypted_salted_ip_address, :last_authentication_timestamp)
    RETURNING uuid, first_authentication_timestamp
"""
QG__validated_ip_address_select_all = "SELECT * FROM validated_ip_address"
QG__validated_ip_address_select = "SELECT * FROM validated_ip_address WHERE uuid = :uuid"
QG__validated_ip_address_update = """
    UPDATE
        validated_ip_address
    SET
        account = COALESCE(:account, account),
        encrypted_salted_ip_address = COALESCE(:encrypted_salted_ip_address, encrypted_salted_ip_address),
        first_authentication_timestamp = COALESCE(:first_authentication_timestamp, first_authentication_timestamp),
        last_authentication_timestamp = COALESCE(:last_authentication_timestamp, last_authentication_timestamp)
    WHERE
        uuid = :uuid
"""


# Generated methods for table 'validated_ip_address'
def validated_ip_address__delete(
    connection: DBInterface,
    uuid
):
    execute_supplied_statement(
        connection, QG__validated_ip_address_delete,
        {
            # Key Fields
            KG__validated_ip_address__uuid: uuid
        }
    )


def validated_ip_address__update(
    connection: DBInterface, encryption_key: bytes,
    uuid,
    account=None, encrypted_salted_ip_address=None, first_authentication_timestamp=None,
    last_authentication_timestamp=None,
    encryption_salts=None
):
    execute_supplied_statement(
        connection, QG__validated_ip_address_update,
        {
            # Key Fields
            KG__validated_ip_address__uuid: uuid,

            # None Key Fields
            KG__validated_ip_address__account: account,
            KG__validated_ip_address__encrypted_salted_ip_address: encrypted_salted_ip_address,
            KG__validated_ip_address__first_authentication_timestamp: first_authentication_timestamp,
            KG__validated_ip_address__last_authentication_timestamp: last_authentication_timestamp
        }, encryption_key=encryption_key, encryption_salts=encryption_salts, encrypt_parameters=[
            KG__validated_ip_address__encrypted_salted_ip_address
        ]
    )


def validated_ip_address__select(
    connection: DBInterface, encryption_key: bytes,
    uuid,
    singleton_code: int = None, singleton_message: str = "validated_ip_address does not exist"
):
    return execute_supplied_statement_singleton(
        connection, QG__validated_ip_address_select, 
        {
            # Key Fields
            KG__validated_ip_address__uuid: uuid
        }, encryption_key=encryption_key, decrypt_columns=[
            KG__validated_ip_address__encrypted_salted_ip_address
        ],
        as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


def validated_ip_address__select_all(
    connection: DBInterface, encryption_key: bytes
):
    return execute_supplied_statement(
        connection, QG__validated_ip_address_select_all, encryption_key=encryption_key, decrypt_columns=[
            KG__validated_ip_address__encrypted_salted_ip_address
        ],
        as_objects=True
    )


def validated_ip_address__insert(
    connection: DBInterface, encryption_key: bytes,
    account, encrypted_salted_ip_address, last_authentication_timestamp,
    encryption_salts=None
):
    return execute_supplied_statement_singleton(
        connection, QG__validated_ip_address_insert,
        {
            KG__validated_ip_address__account: account,
            KG__validated_ip_address__encrypted_salted_ip_address: encrypted_salted_ip_address,
            KG__validated_ip_address__last_authentication_timestamp: last_authentication_timestamp
        }, encryption_key=encryption_key, encryption_salts=encryption_salts, encrypt_parameters=[
            KG__validated_ip_address__encrypted_salted_ip_address
        ], as_objects=True
    )


# Generated keys for table 'security_event'
KG__security_event__application = "application"
KG__security_event__event_lock = "event_lock"
KG__security_event__creation_timestamp = "creation_timestamp"
KG__security_event__wrong_key_attempt_count = "wrong_key_attempt_count"
KG__security_event__email_template = "email_template"
KG__security_event__account = "account"
KG__security_event__unlock_key = "unlock_key"
KG__security_event__unlock_code = "unlock_code"
KG__security_event__unlock_timestamp = "unlock_timestamp"

# Generated queries for table 'security_event'
QG__security_event_delete = "DELETE FROM security_event WHERE application = :application AND event_lock = :event_lock"
QG__security_event_insert = """
    INSERT INTO security_event (application, email_template, account,
        unlock_code, unlock_timestamp)
    VALUES (:application, :email_template, :account,
        :unlock_code, :unlock_timestamp)
    RETURNING event_lock, creation_timestamp, wrong_key_attempt_count,
        unlock_key
"""
QG__security_event_select_all = "SELECT * FROM security_event"
QG__security_event_select = "SELECT * FROM security_event WHERE application = :application AND event_lock = :event_lock"
QG__security_event_update = """
    UPDATE
        security_event
    SET
        creation_timestamp = COALESCE(:creation_timestamp, creation_timestamp),
        wrong_key_attempt_count = COALESCE(:wrong_key_attempt_count, wrong_key_attempt_count),
        email_template = COALESCE(:email_template, email_template),
        account = COALESCE(:account, account),
        unlock_key = COALESCE(:unlock_key, unlock_key),
        unlock_code = COALESCE(:unlock_code, unlock_code),
        unlock_timestamp = COALESCE(:unlock_timestamp, unlock_timestamp)
    WHERE
        application = :application AND event_lock = :event_lock
"""


# Generated methods for table 'security_event'
def security_event__delete(
    connection: DBInterface,
    application, event_lock
):
    execute_supplied_statement(
        connection, QG__security_event_delete,
        {
            # Key Fields
            KG__security_event__application: application,
            KG__security_event__event_lock: event_lock
        }
    )


def security_event__update(
    connection: DBInterface,
    application, event_lock,
    creation_timestamp=None, wrong_key_attempt_count=None, email_template=None,
    account=None, unlock_key=None, unlock_code=None,
    unlock_timestamp=None
):
    execute_supplied_statement(
        connection, QG__security_event_update,
        {
            # Key Fields
            KG__security_event__application: application,
            KG__security_event__event_lock: event_lock,

            # None Key Fields
            KG__security_event__creation_timestamp: creation_timestamp,
            KG__security_event__wrong_key_attempt_count: wrong_key_attempt_count,
            KG__security_event__email_template: email_template,
            KG__security_event__account: account,
            KG__security_event__unlock_key: unlock_key,
            KG__security_event__unlock_code: unlock_code,
            KG__security_event__unlock_timestamp: unlock_timestamp
        }
    )


def security_event__select(
    connection: DBInterface,
    application, event_lock,
    singleton_code: int = None, singleton_message: str = "security_event does not exist"
):
    return execute_supplied_statement_singleton(
        connection, QG__security_event_select, 
        {
            # Key Fields
            KG__security_event__application: application,
            KG__security_event__event_lock: event_lock
        },
        as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


def security_event__select_all(
    connection: DBInterface
):
    return execute_supplied_statement(
        connection, QG__security_event_select_all,
        as_objects=True
    )


def security_event__insert(
    connection: DBInterface,
    application, email_template, account,
    unlock_code,
    unlock_timestamp=None
):
    return execute_supplied_statement_singleton(
        connection, QG__security_event_insert,
        {
            KG__security_event__application: application,
            KG__security_event__email_template: email_template,
            KG__security_event__account: account,
            KG__security_event__unlock_code: unlock_code,
            KG__security_event__unlock_timestamp: unlock_timestamp
        }, as_objects=True
    )



from .exception_queries import *  # Do not remove as others refer to this import

QUERY__setup_jaaql_role = "SELECT setup_jaaql_role();"
QUERY__setup_jaaql_role_with_password = "SELECT setup_jaaql_role_with_password(%(password)s);"



HEADER__security = "Authentication-Token"
HEADER__security_bypass = "Authentication-Token-Bypass"

SQL__where = "WHERE"
SQL__order_by = "ORDER BY"
SQL__limit = "LIMIT"
SQL__offset = "OFFSET"
SQL__asc = "ASC"
SQL__desc = "DESC"
SQL__and = "AND"
SQL__or = "OR"
SQL__like = "LIKE"
SQL__eq = "="
SQL__paren_open = "("
SQL__paren_close = ")"
SQL__single_quote = "'"
SQL__double_quote = '"'
SQL__lower = "lower"
SQL__varchar_cast = "::varchar"
SQL__is_null = "is null"

KEY__username = "username"
KEY__password = "password"
KEY__attach_as = "attach_as"
KEY__ip_address = "ip_address"
KEY__created = "created"
KEY__ip_id = "ip_id"
KEY__application = "application"
KEY__schema = "schema"
KEY__database = "database"
KEY__role = "role"
KEY__read_only = "read_only"
KEY__install_key = "install_key"
KEY__jaaql_password = "jaaql_password"
KEY__super_db_password = "super_db_password"
KEY__old_password = "old_password"
KEY__position = "position"
KEY__file = "file"
KEY__error = "error"
KEY__error_row_number = "row_number"
KEY__error_index = "index"
KEY__error_query = "query"
KEY__error_set = "set"
KEY__allow_uninstall = "allow_uninstall"
KEY__sign_up_template = "sign_up"
KEY__already_signed_up_template = "already_signed_up"
KEY__reset_password_template = "reset_password"
KEY__unregistered_user_reset_password_template = "unregistered_user_reset_password"
KEY__parameters = "parameters"
KEY__security_key = "security_key"
KEY__id = "id"

SEPARATOR__comma_space = ", "
SEPARATOR__comma = ","
SEPARATOR__space = " "

JAAQL__arg_marker = ":"

DIR__config = "config"
DIR__www = "www"
DIR__render_template = "rendered_documents"

FILE__config = "config.ini"
FILE__canned_queries = "canned_queries.jsql"

ENCODING__utf = "UTF-8"
ENCODING__ascii = "ascii"
EXAMPLE__jwt = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.e30.yXvILkvUUCBqAFlAv6wQ1Q-QRAjfe3eSosO949U73Vo"

VAULT_KEY__db_crypt_key = "db_crypt_key"
VAULT_KEY__db_repeatable_salt = "repeatable_salt"
VAULT_KEY__jaaql_local_access_key = "jaaql_access_key"
VAULT_KEY__super_db_credentials = "super_db_credentials"
VAULT_KEY__jaaql_lookup_connection = "jaaql_lookup_connection"
VAULT_KEY__allow_jaaql_uninstall = "Allow jaaql uninstall"
VAULT_KEY__jaaql_db_password = "Jaaql DB Password"

# Used for re-installation
VAULT_KEY__db_connection_string = "db_connection_string"
VAULT_KEY__jaaql_password = "jaaql_password"
VAULT_KEY__super_db_password = "super_db_password"

ENVIRON__vault_key = "JAAQL_VAULT_PASSWORD"
ENVIRON__local_install = "JAAQL_LOCAL_INSTALL"
ENVIRON__jaaql_profiling = "JAAQL_PROFILING"
ENVIRON__install_path = "INSTALL_PATH"
ENVIRON__sentinel_url = "SENTINEL_URL"
ENVIRON__canned_queries = "CANNED_QUERIES"

EMAIL_PARAM__unlock_key = "JAAQL__UNLOCK_KEY"
EMAIL_PARAM__unlock_code = "JAAQL__UNLOCK_CODE"
EMAIL_PARAM__app_url = "JAAQL__APP_URL"
EMAIL_PARAM__email_address = "JAAQL__EMAIL_ADDRESS"

CONFIG_KEY__server = "SERVER"
CONFIG_KEY_SERVER__port = "port"

JWT_PURPOSE__oauth = "oauth"
JWT_PURPOSE__pre_auth = "pre_auth"
JWT_PURPOSE__connection = "connection"

ERR__invalid_token = "Invalid token!"
ERR__not_yet_installed = "JAAQL has not yet been installed!"
ERR__user_public = "Cannot perform this action on a public user!"
ERR__too_many_signup_attempts = "Too many signup attempts"
ERR__too_many_reset_requests = "Too many reset requests"
ERR__too_many_code_attempts = "Code disabled due to too many incorrect attempts. Please use the link in the email"
ERR__document_still_rendering = "Document still rendering"
ERR__document_id_not_found = "Document id not found"
ERR__unlock_code_expired = "The short unlock code has expired. Please use the long link found in your email"
ERR__invalid_lock = "Either security event does not exist, has already been used, has expired"
ERR__incorrect_lock_code = "Incorrect lock code"

PG_ENV__password = "POSTGRES_PASSWORD"

HTML__base64_png = "data:image/png;base64,"
FORMAT__png = "png"

NODE__host_node = "host"
DB__jaaql = "jaaql"
DB__postgres = "postgres"

PORT__ems = 6061

ENDPOINT__send_email = "/send-email"
ENDPOINT__internal_applications = "/internal/applications"
ENDPOINT__internal_templates = "/internal/emails/templates"
ENDPOINT__internal_accounts = "/internal/emails/accounts"
ENDPOINT__is_alive = "/internal/is-alive"
ENDPOINT__jaaql_emails = "/emails"
ENDPOINT__report_sentinel_error = "/sentinel/reporting/error"
ENDPOINT__install = "/internal/install"

CONFIG__default = "Default config"
CONFIG__default_desc = "Default config description"
DATASET__default = "Default dataset"
DATASET__default_desc = "Default dataset description"

ARTIFACTS_DEFAULT_DIRECTORY = "artifacts"

USERNAME__jaaql = "jaaql"
USERNAME__super_db = "super_db"
USERNAME__anonymous = "anonymous"
PASSWORD__anonymous = "jaaql_public_password"
ROLE__jaaql = "jaaql"
ROLE__postgres = "postgres"

PROTOCOL__postgres = "postgresql://"

VERSION = "4.6.0"




"""
This script was generated from jaaql.exceptions.fxls at 12/03/2023, 15:27:49
"""

from jaaql.utilities.crypt_utils import get_repeatable_salt
from .generated_queries import *


def add_account_password(
    connection: DBInterface, encryption_key: bytes, vault_repeatable_salt: str,
    account__id, account_password__hash
):
    account_password__uuid = account_password__insert(
        connection, encryption_key, account__id, 
        account_password__hash, 
        encryption_salts={KG__account_password__hash: get_repeatable_salt(vault_repeatable_salt, account__id)}
    )[KG__account_password__uuid]

    account__update(
        connection, encryption_key, account__id,
        most_recent_password=account_password__uuid
    )


QUERY__fetch_most_recent_password = """
    SELECT
        A.id,
        P.uuid,
        P.hash
    FROM account_password P
    INNER JOIN account A ON P.uuid = A.most_recent_password
    WHERE
        A.id = :id
"""


def fetch_most_recent_password(
    connection: DBInterface, encryption_key: bytes, account__id
):
    return execute_supplied_statement_singleton(
        connection, QUERY__fetch_most_recent_password, {KG__account__id: account__id},
        decrypt_columns=[KG__account_password__hash], encryption_key=encryption_key, as_objects=True
    )[KG__account_password__hash]


QUERY__fetch_most_recent_password_from_username = """
    SELECT
        A.id,
        P.uuid,
        P.hash
    FROM account_password P
    INNER JOIN account A ON P.uuid = A.most_recent_password
    WHERE
        A.username = :username
"""


def fetch_most_recent_password_from_username(
    connection: DBInterface, encryption_key: bytes, vault_repeatable_salt: str,
    account__username, singleton_code: int = None, singleton_message: str = None
):
    return execute_supplied_statement_singleton(
        connection, QUERY__fetch_most_recent_password_from_username, {KG__account__username: account__username},
        encryption_key=encryption_key, encrypt_parameters=[KG__account__username], encryption_salts={
            KG__account__username: get_repeatable_salt(vault_repeatable_salt)
        }, as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


QUERY__fetch_account_from_username = "SELECT * FROM account WHERE username = :username"


def fetch_account_from_username(
    connection: DBInterface, encryption_key: bytes, vault_repeatable_salt: str,
    username, singleton_code: int = None, singleton_message: str = None
):
    return execute_supplied_statement_singleton(
        connection, QUERY__fetch_account_from_username, {KG__account__username: username},
        encryption_key=encryption_key, encrypt_parameters=[KG__account__username], encryption_salts={
            KG__account__username: get_repeatable_salt(vault_repeatable_salt)
        }, as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )


QUERY__create_account = "SELECT create_account(:username, :attach_as, :already_exists, :is_the_anonymous_user) as account_id"
KEY__attach_as = "attach_as"
KEY__already_exists = "already_exists"
KEY__username = "username"
KEY__is_the_anonymous_user = "is_the_anonymous_user"
KEY__account_id = "account_id"


def create_account(
    connection: DBInterface, encryption_key: bytes, vault_repeatable_salt: str,
    username, attach_as=None, already_exists=False, is_the_anonymous_user=False
):
    return execute_supplied_statement_singleton(
        connection, QUERY__create_account, {
            KEY__username: username,
            KEY__attach_as: attach_as,
            KEY__already_exists: already_exists,
            KEY__is_the_anonymous_user: is_the_anonymous_user
        }, encryption_salts={
            KG__account__username: get_repeatable_salt(vault_repeatable_salt)
        }, as_objects=True, encryption_key=encryption_key, encrypt_parameters=[KG__account__username]
    )[KEY__account_id]


QUERY__validate_most_recent_password = "SELECT * FROM account WHERE id = :id AND most_recent_password = :most_recent_password"


def validate_is_most_recent_password(
    connection: DBInterface, account,
    uuid, singleton_code: int = None, singleton_message: str = None
):
    return execute_supplied_statement_singleton(
        connection, QUERY__validate_most_recent_password, {
            KG__account__id: account,
            KG__account__most_recent_password: uuid
        }, singleton_code=singleton_code, singleton_message=singleton_message, skip_commit=True
    )


QUERY__exists_matching_validated_ip_address = "SELECT COUNT(*) FROM validated_ip_address WHERE encrypted_salted_ip_address = :encrypted_salted_ip_address"
KEY__count = "count"


def exists_matching_validated_ip_address(
    connection: DBInterface, encrypted_salted_ip_address
):
    return execute_supplied_statement_singleton(
        connection, QUERY__exists_matching_validated_ip_address, {
            KG__validated_ip_address__encrypted_salted_ip_address: encrypted_salted_ip_address
        }, as_objects=True
    )[KEY__count] == 1


QUERY___add_or_update_validated_ip_address = "INSERT INTO validated_ip_address (account, encrypted_salted_ip_address) VALUES (:account, :encrypted_salted_ip_address) ON CONFLICT ON CONSTRAINT validated_ip_address_encrypted_salted_ip_address_key DO UPDATE SET last_authentication_timestamp = current_timestamp RETURNING uuid as uuid"


QUERY__fetch_application_schemas = "SELECT S.name, S.database, (A.default_schema = S.name) as is_default, A.is_live FROM application_schema S INNER JOIN application A ON A.name = S.application WHERE S.application = :application"
KEY__is_default = "is_default"

QUERY__count_security_events_of_type_in_24hr_window = """
    SELECT
        COUNT(*)
    FROM security_event S
    INNER JOIN email_template E ON E.name = S.email_template AND E.application = S.application
    WHERE E.type IN (:type_one, :type_two) AND account = :account AND (creation_timestamp + interval '24 hour') > current_timestamp
"""

KEY__type_one = "type_one"
KEY__type_two = "type_two"

EMAIL_TYPE__signup = "S"
EMAIL_TYPE__already_signed_up = "A"
EMAIL_TYPE__reset_password = "R"
EMAIL_TYPE__unregistered_password_reset = "U"
EMAIL_TYPE__general = "G"

KEY__key_fits = "key_fits"
QUERY__check_security_event_unlock = """
    UPDATE
        security_event S
    SET
        wrong_key_attempt_count = S.wrong_key_attempt_count + (case when (S.unlock_code = :unlock_code OR S.unlock_key = :unlock_key OR S.wrong_key_attempt_count >= 3) then 0 else 1 end)
    FROM application A
    WHERE
        S.application = A.name AND
        S.event_lock = :event_lock AND S.unlock_timestamp is null AND
        S.creation_timestamp + (A.unlock_code_validity_period || ' seconds')::interval > current_timestamp
    RETURNING S.*, A.unlock_code_validity_period, (S.unlock_code = :unlock_code OR S.unlock_key = :unlock_key) as key_fits
"""


def check_security_event_unlock(
    connection: DBInterface, event_lock, unlock_code,
    unlock_key, singleton_code: int = None, singleton_message: str = None
):
    return execute_supplied_statement_singleton(
        connection, QUERY__check_security_event_unlock, {
            KG__security_event__event_lock: event_lock,
            KG__security_event__unlock_code: unlock_code,
            KG__security_event__unlock_key: unlock_key
        }, as_objects=True, singleton_code=singleton_code, singleton_message=singleton_message
    )



import os
import traceback
import uuid
import time

from jaaql.constants import *
from jaaql.utilities.utils import load_config, await_jaaql_installation, get_jaaql_connection
from jaaql.utilities.utils_no_project_imports import check_allowable_file_path
from email.utils import formatdate, make_msgid
from smtplib import SMTP, SMTPException, SMTPAuthenticationError
from base64 import urlsafe_b64encode as b64e
from typing import Union, List
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from queue import Queue, Empty
import re
import hashlib
import json
import ssl
import sys
import threading
from flask import Flask, jsonify, request
from jaaql.mvc.generated_queries import *
from jaaql.utilities.vault import Vault, DIR__vault
from jaaql.constants import VAULT_KEY__db_crypt_key

ELE__jaaql_filename = "jaaql_filename"

ERR__email_not_found = "Email account not found with name '%s'"
ERR__invalid_call_to_internal_email_service = "Invalid call to internal email service"
ERR__attachment_timeout = "Could not send email! Attachment rendering timed out!"
ERR__attachment_error = "Could not send email! Attachment rendering received error: '%s'"
ERR__attachment_filename = "Attachment did not render in time or could not find the file name for the attachment. Please check template. Expected " \
    "presence of element with id " + ELE__jaaql_filename + " to be present after page has finished loading"

PROTOCOL__imap = "imap"
PROTOCOL__smtp = "smtp"

EMAIL__from = "From"
EMAIL__from_email = "From_Email"
EMAIL__to = "To"
EMAIL__subject = "Subject"
EMAIL__date = "Date"
EMAIL__message_id = "Message-ID"

SPLIT__address = ", "

TIMEOUT__attachment = 60000
TIMEOUT__filename = 15000

REGEX__html_tags = r'<(\/[a-zA-Z]+|(!?[a-zA-Z]+((\s)*((\"?[^(\">)]*\"?)|[a-zA-Z]*\s*=\s*\"[^\"]*\"))*))>'

HTTP_ARG__oauth_token = "oauth_token"


class EmailAttachment:
    def __init__(self, name: str, parameters: dict, template: str, artifact_base: str = None):
        self.name = name
        self.artifact_base = artifact_base
        self.parameters = parameters
        self.id = str(uuid.uuid4())
        self.filename = None
        self.content = None
        self.access_token = None
        self.template = str(template)

    @staticmethod
    def static_format_attached_url(artifact_base_url: str, is_container: bool):
        if not artifact_base_url.startswith("https://") and not artifact_base_url.startswith("http://"):
            if is_container:
                if check_allowable_file_path(artifact_base_url):
                    raise Exception("Illegal artifact base url")

            if artifact_base_url.startswith("file:///"):
                return artifact_base_url
            else:
                return "file://" + os.environ[ENVIRON__install_path] + "/www/" + artifact_base_url
        else:
            return artifact_base_url

    def format_attachment_url(self, artifact_base_url: str, is_container: bool):
        self.artifact_base = EmailAttachment.static_format_attached_url(artifact_base_url, is_container)

    def repr_json(self):
        return dict(name=self.name, parameters=self.parameters, template=self.template, artifact_base=self.artifact_base)

    @staticmethod
    def deserialize(attachment: dict, template: str = None, artiface_base_url: str = None):
        if template is not None:
            return EmailAttachment(attachment[KEY__attachment_name], attachment[KEY__parameters], template, artiface_base_url)
        else:
            return EmailAttachment(attachment[KEY__attachment_name], attachment[KEY__parameters], attachment[KEY__template],
                                   attachment[KEY__artifact_base])

    @staticmethod
    def deserialize_list(attachments: list, template: str, artiface_base_url: str):
        if attachments is None:
            return None
        else:
            return [EmailAttachment.deserialize(attachment, template, artiface_base_url) for attachment in attachments]

    def encode_filename(self):
        return b64e(self.filename.encode(ENCODING__ascii)).decode(ENCODING__ascii)

    def encode_content(self):
        return b64e(self.content).decode(ENCODING__ascii)


TYPE__email_attachments = Union[EmailAttachment, List[EmailAttachment]]


class Email:
    def __init__(self, application: str, template: str, dispatcher: str, to: Union[str, List[str]],
                 subject: str, body: str, attachments: TYPE__email_attachments = None, recipient_names: Union[str, List[str]] = None,
                 is_html: bool = True, attachment_access_token: str = None, override_send_name: str = None):
        self.application = application
        self.template = template
        self.dispatcher = dispatcher
        if recipient_names is None:
            recipient_names = to
        if isinstance(to, list) != isinstance(recipient_names, list):
            raise Exception(ERR__invalid_call_to_internal_email_service)
        if not isinstance(to, list):
            to = [to]
            recipient_names = [recipient_names]
        if len(to) != len(recipient_names):
            raise Exception(ERR__invalid_call_to_internal_email_service)
        recipient_names = [to[idx] if itm is None else itm for idx, itm in zip(range(len(recipient_names)), recipient_names)]
        self.to = to
        self.recipient_names = recipient_names
        self.subject = subject
        self.body = body
        self.attachments = attachments
        self.override_send_name = override_send_name
        if not isinstance(self.attachments, List) and self.attachments is not None:
            self.attachments = [attachments]
        self.is_html = is_html
        self.attachment_access_token = attachment_access_token

    def repr_json(self):
        return dict(application=self.application, template=self.template, dispatcher=self.dispatcher,
                    to=self.to, recipient_names=self.recipient_names, subject=self.subject, body=self.body,
                    attachments=[attachment.repr_json() for attachment in self.attachments] if self.attachments is not None else None,
                    is_html=self.is_html, attachment_access_token=self.attachment_access_token, override_send_name=self.override_send_name)

    @staticmethod
    def deserialize(email: dict):
        attachments = None
        if email["attachments"] is not None:
            attachments = [EmailAttachment.deserialize(attachment) for attachment in email["attachments"]]
        return Email(email["application"], email["template"], email["dispatcher"], email["to"],
                     email["subject"], email["body"], attachments, email["recipient_names"], email["is_html"], email["attachment_access_token"],
                     email["override_send_name"])


class ParameterisedLock:
    def __init__(self):
        self.namespace_lock = threading.Lock()
        self._lock_dict = {}

    def acquire_lock(self, key: str, blocking: bool = False):
        if key not in self._lock_dict:
            with self.namespace_lock:
                if key not in self._lock_dict:
                    self._lock_dict[key] = threading.Lock()

        return self._lock_dict[key].acquire(blocking=blocking)

    def release_lock(self, key: str):
        self._lock_dict[key].release()

    def check_lock_held(self, key: str):
        if key not in self._lock_dict:
            return False
        else:
            return self._lock_dict[key].locked()


class ThreadsafeQueueInitialiser:
    def __init__(self):
        self.parameterised_lock = ParameterisedLock()
        self.queues = {}

    def fetch_queue(self, key: str):
        if key not in self.queues:
            self.parameterised_lock.acquire_lock(key, True)
            if key not in self.queues:
                self.queues[key] = Queue()
            self.parameterised_lock.release_lock(key)
        return self.queues[key]


class EmailManagerService:

    def __init__(self, connection: DBInterface, db_crypt_key: bytes, is_gunicorn: bool):
        self.connection = connection
        self.db_crypt_key = db_crypt_key
        self.is_gunicorn = is_gunicorn
        self.dispatchers = {}
        self.dispatcher_blacklist = {}
        self.parameterised_lock = ParameterisedLock()
        self.threadsafe_queue_initialiser = ThreadsafeQueueInitialiser()
        threading.Thread(target=self.dispatcher_manager_thread, daemon=True).start()

    def construct_message(self, dispatcher_info: dict, email: Email) -> (str, MIMEMultipart):
        send_body = email.body

        message = MIMEMultipart("alternative")

        if email.is_html:
            message.attach(MIMEText(re.sub(REGEX__html_tags, "", send_body).strip()))
            message.attach(MIMEText(send_body, 'html'))
        else:
            message.attach(MIMEText(send_body, 'plain'))

        message_final = MIMEMultipart('mixed')
        message_final.attach(message)
        message_final[EMAIL__to] = SPLIT__address.join(email.to)
        send_name = dispatcher_info[KG__email_dispatcher__display_name]
        if email.override_send_name is not None:
            send_name = email.override_send_name
        message_final[EMAIL__from] = send_name + " <" + dispatcher_info[KG__email_dispatcher__username] + ">"
        message_final[EMAIL__subject] = email.subject
        message_final[EMAIL__date] = formatdate()
        message_final[EMAIL__message_id] = make_msgid()

        if email.attachments is not None:
            email_attachments = email.attachments
            for attachment in email_attachments:
                pass  # TODO set up attachments

        return send_body, message_final

    def is_connected(self, conn: SMTP):
        try:
            status = conn.noop()[0]
        except Exception:
            status = -1
        return True if status == 250 else False

    def try_close_conn(self, conn):
        try:
            conn.close()
        except:
            pass

    def fetch_conn(self, dispatcher_key: str, host: str, port: int, username: str, password: str):
        if not self.is_gunicorn:
            return None

        conn = SMTP(host, port=port, timeout=30)
        if conn is None:
            raise Exception("Could not connect to email dispatcher '%s' with address '%s:%d'" % dispatcher_key, host, port)

        context = ssl.SSLContext(ssl.PROTOCOL_TLS)

        conn.starttls(context=context)
        conn.login(username, password)

        return conn

    def form_dispatcher_key(self, application: str, name: str):
        return application + ":" + name

    def fetch_dispatchers_from_db(self):
        return {
            self.form_dispatcher_key(dispatcher[KG__email_dispatcher__application], dispatcher[KG__email_dispatcher__name]): (
                dispatcher,
                self.get_dispatcher_queue_from_key(self.form_dispatcher_key(dispatcher[KG__email_dispatcher__application],
                                                                            dispatcher[KG__email_dispatcher__name]))
            )
            for dispatcher in email_dispatcher__select_all(self.connection, self.db_crypt_key)
        }

    def update_dispatchers(self):
        new_dispatchers = self.fetch_dispatchers_from_db()
        potentially_require_threads = {key: val for key, val in new_dispatchers.items() if not self.parameterised_lock.check_lock_held(key)}
        self.dispatchers = new_dispatchers
        for dispatcher_key, dispatcher_info in potentially_require_threads.items():
            threading.Thread(target=self.dispatcher_thread, args=[dispatcher_key, dispatcher_info]).start()

    def get_dispatcher_queue_from_key(self, dispatcher_key: str):
        return self.dispatchers.get(dispatcher_key, (None, self.threadsafe_queue_initialiser.fetch_queue(dispatcher_key)))[1]

    def get_dispatcher_info_from_key(self, dispatcher_key: str):
        return self.dispatchers.get(dispatcher_key, (None, None))[0]

    def dispatcher_manager_thread(self):
        printed = False
        while True:
            try:
                self.update_dispatchers()
                if printed:
                    print("Successfully refreshed dispatchers after downtime")
                printed = False
            except Exception:
                if not printed:
                    print("Could not update dispatchers. Is db disconnected?")
                    traceback.print_exc()
                printed = True

            time.sleep(15)

    def send_email_with_connection(self, conn: SMTP, email: Email, dispatcher_key: str, dispatcher_info: dict):
        send_to = email.to
        send_recipients = email.recipient_names
        whitelist = dispatcher_info[KG__email_dispatcher__whitelist]

        if whitelist is not None and whitelist != "":
            send_to = []
            send_recipients = []
            for to, recipient in zip(email.to, email.recipient_names):
                if "@" not in to:
                    raise Exception("Email address '%s' was invalid" % to)
                start_len = len(send_to)

                for check in whitelist.split(","):
                    check = check.strip()
                    domain = check.split("@")[1].lower()

                    to_domain = to.strip().split("@")[1].lower()
                    if check.startswith("*@"):
                        if domain == to_domain:
                            send_to.append(to)
                            send_recipients.append(recipient)
                            break
                    elif domain == to_domain:
                        person = check.split("@")[0].lower()
                        to_person = to.strip().split("@")[0].lower().split("+")[0]
                        if person == to_person:
                            send_to.append(to)
                            send_recipients.append(recipient)
                            break

                if len(send_to) == start_len:
                    print("Address '%s' rejected by whitelist for account '%s'" % (to, dispatcher_key), file=sys.stderr)

        if len(send_to) == 0:
            return

        email.to = send_to
        email.recipient_names = send_recipients
        formatted_body, to_send = self.construct_message(dispatcher_info, email)

        if self.is_gunicorn:
            conn.send_message(to_send, dispatcher_info[KG__email_dispatcher__display_name], send_to)
        else:
            print("==============================================")
            print("Sending Email from '" + dispatcher_key + "'")
            print(email.subject)
            print(email.body)
            print(dispatcher_info[KG__email_dispatcher__display_name])
            print(send_to)
            print("==============================================")

    def hash_dispatcher_info(self, dispatcher_info):
        return hashlib.md5(json.dumps(dispatcher_info, default=str).encode()).hexdigest()

    def dispatcher_thread(self, dispatcher_key: str, dispatcher_info: dict):
        did_acquire = self.parameterised_lock.acquire_lock(dispatcher_key)
        if not did_acquire:
            return  # Did not acquire the thread lock. Another thread is operating

        queue = self.get_dispatcher_queue_from_key(dispatcher_key)
        email = None

        if self.dispatcher_blacklist.get(dispatcher_key, None) != self.hash_dispatcher_info(dispatcher_info):
            if dispatcher_key in self.dispatcher_blacklist:
                self.dispatcher_blacklist.pop(dispatcher_key)

            try:
                conn = None
                while (dispatcher_key in self.dispatchers or queue.qsize() != 0) and dispatcher_key not in self.dispatcher_blacklist:
                    try:
                        # The dispatcher has been removed from the database AND all emails in the queue have been removed
                        email = queue.get(timeout=10)
                        new_dispatcher_info = self.get_dispatcher_info_from_key(dispatcher_key)
                        if new_dispatcher_info is not None:
                            dispatcher_info = new_dispatcher_info
                        else:
                            new_dispatcher_info = dispatcher_info  # Done for hashing purposes

                        if self.hash_dispatcher_info(dispatcher_info) != self.hash_dispatcher_info(new_dispatcher_info):
                            self.try_close_conn(conn)

                        if not self.is_connected(conn):
                            conn = self.fetch_conn(dispatcher_key, dispatcher_info[KG__email_dispatcher__url],
                                                   dispatcher_info[KG__email_dispatcher__port], dispatcher_info[KG__email_dispatcher__username],
                                                   dispatcher_info[KG__email_dispatcher__password])

                        self.send_email_with_connection(conn, email, dispatcher_key, dispatcher_info)
                    except Empty:
                        pass  # Nothing wrong here. There were simply no emails received in the 10 seconds of queue wait time
                    except SMTPAuthenticationError as ex:
                        raise ex  # Caught in outer loop, we want to break free
                    except SMTPException as ex:
                        print("Issue sending email " + str(ex), file=sys.stderr)

            except SMTPAuthenticationError:
                queue.put(email)
                print("Incorrect credentials for email dispatcher with key '%s'" % dispatcher_key, file=sys.stderr)
                self.dispatcher_blacklist[dispatcher_key] = self.hash_dispatcher_info(dispatcher_info)
            except Exception:
                print("Unhandled email dispatcher exception", file=sys.stderr)
                traceback.print_exc()

        self.parameterised_lock.release_lock(dispatcher_key)

    def send_email(self, email: Email):
        dispatcher_key = self.form_dispatcher_key(email.application, email.dispatcher)
        if dispatcher_key not in self.dispatchers:
            self.update_dispatchers()
            if dispatcher_key not in self.dispatchers:
                raise Exception("Unable to find dispatcher with key '" + dispatcher_key + "'")
        self.get_dispatcher_queue_from_key(dispatcher_key).put(email)


def create_app(ems: EmailManagerService):

    app = Flask(__name__, instance_relative_config=True)

    @app.route(ENDPOINT__send_email, methods=["POST"])
    def send_email():
        ems.send_email(Email.deserialize(request.json))

        return jsonify("OK")

    @app.route("/", methods=["GET"])
    def is_alive():
        return jsonify("OK")

    return app


def create_flask_app(vault_key=None, is_gunicorn: bool = False):
    config = load_config(is_gunicorn)
    await_jaaql_installation(config, is_gunicorn)
    vault = Vault(vault_key, DIR__vault)
    jaaql_lookup_connection = get_jaaql_connection(config, vault)
    db_crypt_key = vault.get_obj(VAULT_KEY__db_crypt_key).encode(ENCODING__ascii)

    flask_app = create_app(EmailManagerService(jaaql_lookup_connection, db_crypt_key, is_gunicorn))
    print("Created email manager app host, running flask", file=sys.stderr)
    flask_app.run(port=PORT__ems, host="0.0.0.0", threaded=True)



import re
import uuid

from jaaql.db.db_pg_interface import DBPGInterface
from jaaql.documentation.documentation_public import KEY__oauth_token
from jaaql.mvc.base_model import BaseJAAQLModel, VAULT_KEY__jwt_crypt_key
from jaaql.exceptions.http_status_exception import HttpStatusException, HTTPStatus, ERR__already_installed
from os.path import join
from jaaql.constants import *
from jaaql.mvc.response import JAAQLResponse
from jaaql.interpreter.interpret_jaaql import InterpretJAAQL, ASSERT_one, KEY_assert, KEY_query, KEY_parameters
from jaaql.utilities.utils import get_jaaql_root
from jaaql.db.db_utils import create_interface, jaaql__encrypt
from jaaql.utilities import crypt_utils
import threading
from datetime import datetime, timedelta
from jaaql.mvc.handmade_queries import *
from jwt.utils import base64url_decode
import json
import os
from queue import Queue
import subprocess
import time
import random

REGEX__object_name = r'^[0-9a-zA-Z_]{1,63}$'

ERR__invalid_object_name = "Object name '%s' is invalid. Must match regex: " + REGEX__object_name
ERR__refresh_expired = "Token too old to be used for refresh. Please authenticate again"
ERR__incorrect_install_key = "Incorrect install key!"
ERR__invalid_level = "Invalid level!"
ERR__incorrect_credentials = "Incorrect credentials!"
ERR__email_template_not_installed = "Either email template does not exist"
ERR__lacking_permissions = "Only an administrator can perform this action!"
ERR__schema_invalid = "Schema invalid!"
ERR__cant_send_attachments = "Cannot send attachments to other people"
ERR__keep_alive_failed = "Keep alive failed"
ERR__template_not_signup = "Sign up template does not have the correct type"
ERR__template_not_already = "Already signed up template does not have the correct type"
ERR__template_not_reset = "Reset template does not have the correct type"
ERR__template_not_unregistered = "Unregistered template does not have the correct type"
ERR__unexpected_parameters = "Email parameters were not expected"
ERR__expected_parameters = "Email parameters were expected"
ERR__unexpected_validation_column = "Unexpected column in the input parameters '%s'"
ERR__data_validation_table_no_primary = "Data validation table has no primary key"
ERR__cant_find_sign_up = "Cannot locate sign up with key. The key is either incorrect, has expired or has not been activated with the emailed code"
ERR__invite_code_expired = "Invite code expired. Please use the link within the email"
ERR__incorrect_invite_code = "Incorrect invite code"
ERR__template_reset_password = "You cannot use this template to reset your password"
ERR__cant_find_reset = "Cannot locate reset with key. The key is either incorrect, has expired or has not been activated with the emailed code"
ERR__reset_code_expired = "Reset code expired. Please use the link within the email"
ERR__incorrect_reset_code = "Incorrect reset code"
ERR__as_attachment_unexpected = "Input 'as_attachment' unexpected as document is returned as a file link"
ERR__document_created_file = "Document is a file, cannot be downloaded in this way"

PG__default_connection_string = "postgresql://postgres:%s@localhost:5432/jaaql"
DIR__scripts = "scripts"

MODIFIER__allow_conflicts = " ON CONFLICT DO NOTHING"

ATTR__version = "version"
ATTR__count = "count"
ATTR__the_user = "the_user"
ATTR__data_lookup_json = "data_lookup_json"
ATTR__activated = "activated"
ATTR__code_attempts = "code_attempts"
ATTR__closed = "closed"
ATTR__created = "created"
ATTR__expiry_code_ms = "code_expiry_ms"
ATTR__used_key_a = "used_key_a"

SQL__err_duplicate = "duplicate key value violates unique constraint"

RESEND__invite_max = 3
RESEND__reset_max = 2
RESET__not_started = 0
RESET__started = 1
RESET__completed = 2

CODE__letters = "ABCDEFGHIJKLMNPQRSTUVWXYZ"
CODE__invite_length = 6
CODE__max_attempts = 3
CODE__alphanumeric = "ABCDEFGHIJKLMNPQRSTUVWXYZ123456789"
CODE__reset_length = 8

SIGNUP__not_started = 0
SIGNUP__started = 1
SIGNUP__already_registered = 2
SIGNUP__completed = 3


class JAAQLModel(BaseJAAQLModel):
    VERIFICATION_QUEUE = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def redeploy(self, connection: DBInterface):
        raise HttpStatusException("Not yet implemented", response_code=HTTPStatus.NOT_IMPLEMENTED)

    def create_account_with_potential_password(self, connection: DBInterface, username: str, attach_as: str = None, password: str = None,
                                               already_exists: bool = False, is_the_anonymous_user: bool = False):
        account_id = create_account(connection, self.get_db_crypt_key(), self.get_vault_repeatable_salt(),
                                    username, attach_as, already_exists,
                                    is_the_anonymous_user)

        if password:
            self.add_account_password(account_id, password)

        return account_id

    def is_super_admin(self, connection: DBInterface):
        res = execute_supplied_statement_singleton(connection, "select usesuper from pg_user where usename = CURRENT_USER",
                                                   as_objects=True)["usesuper"]
        if not res:
            raise HttpStatusException("You do not have super user privileges")

    def clean(self, connection: DBInterface):
        self.is_super_admin(connection)

        if self.is_container:
            if not self.vault.get_obj(VAULT_KEY__allow_jaaql_uninstall):
                raise HttpStatusException("JAAQL not permitted to uninstall itself")

            subprocess.call("./pg_reboot.sh", cwd="/")
        else:
            subprocess.call("docker kill jaaql_pg")
            subprocess.call("docker rm jaaql_pg")
            subprocess.Popen("docker run --name jaaql_pg -p 5434:5432 jaaql/jaaql_pg", start_new_session=True)
            time.sleep(7.5)

        DBPGInterface.close_all_pools()
        self.jaaql_lookup_connection = None
        self.install_key = str(uuid.uuid4())
        self.install(
            self.vault.get_obj(VAULT_KEY__db_connection_string),
            self.vault.get_obj(VAULT_KEY__jaaql_password),
            self.vault.get_obj(VAULT_KEY__super_db_password),
            self.install_key,
            True,
            False,
            self.vault.get_obj(VAULT_KEY__jaaql_db_password)
        )

    def add_account_password(self, account_id: str, password: str):
        crypt_utils.validate_password(password)
        salt = self.get_repeatable_salt(account_id)
        add_account_password(
            self.jaaql_lookup_connection, self.get_db_crypt_key(), self.get_vault_repeatable_salt(),
            account_id, crypt_utils.hash_password(password, salt)
        )

    def verify_current_password(self, account_id: str, password: str):
        most_recent_password = fetch_most_recent_password(self.jaaql_lookup_connection, self.get_db_crypt_key(), account_id)
        if not crypt_utils.verify_password_hash(most_recent_password, password, salt=self.get_repeatable_salt(account_id)):
            raise HttpStatusException(ERR__incorrect_credentials)

    def add_my_account_password(self, account_id: str, username: str, ip_address: str, old_password: str, password: str):
        self.verify_current_password(account_id, old_password)
        self.add_account_password(account_id, password)
        return self.get_auth_token(password=password, ip_address=ip_address, username=username)

    def is_installed(self, response: JAAQLResponse):
        if not self.has_installed:
            response.response_code = HTTPStatus.UNPROCESSABLE_ENTITY
            return ERR__not_yet_installed

    def is_alive(self):
        version = execute_supplied_statement_singleton(self.jaaql_lookup_connection, "SELECT version() as version;", as_objects=True)[ATTR__version]
        if "PostgreSQL " not in version:
            raise HttpStatusException(ERR__keep_alive_failed)

    def install(self, db_connection_string: str, jaaql_password: str, super_db_password: str, install_key: str, allow_uninstall: bool,
                do_reboot: bool = True, jaaql_db_password: str = None):
        if self.jaaql_lookup_connection is None:
            if allow_uninstall or not self.is_container:
                self.vault.insert_obj(VAULT_KEY__allow_jaaql_uninstall, True)
                self.vault.insert_obj(VAULT_KEY__db_connection_string, db_connection_string)
                self.vault.insert_obj(VAULT_KEY__jaaql_password, jaaql_password)
                self.vault.insert_obj(VAULT_KEY__super_db_password, super_db_password)

            if install_key != self.install_key:
                raise HttpStatusException(ERR__incorrect_install_key, HTTPStatus.UNAUTHORIZED)

            if db_connection_string is None:
                db_connection_string = PG__default_connection_string % os.environ[PG_ENV__password]

            address, port, _, username, db_password = DBInterface.fracture_uri(db_connection_string)

            super_interface = create_interface(self.config, address, port, DB__jaaql, username, db_password)
            conn = super_interface.get_conn()
            super_interface.execute_script_file(conn, join(get_jaaql_root(), DIR__scripts, "01.install_domains.generated.sql"))
            super_interface.execute_script_file(conn, join(get_jaaql_root(), DIR__scripts, "02.install_super_user.exceptions.sql"))
            super_interface.execute_script_file(conn, join(get_jaaql_root(), DIR__scripts, "03.install_super_user.handwritten.sql"))
            if jaaql_db_password is not None:
                super_interface.execute_query(conn, QUERY__setup_jaaql_role_with_password, {KEY__password: jaaql_db_password})
            else:
                resp = super_interface.execute_query(conn, QUERY__setup_jaaql_role)
                jaaql_db_password = resp[1][0][0]
            if allow_uninstall or not self.is_container:
                self.vault.insert_obj(VAULT_KEY__jaaql_db_password, jaaql_db_password)
            super_interface.commit(conn)
            super_interface.put_conn(conn)
            super_interface.close()

            self.jaaql_lookup_connection = create_interface(self.config, address, port, DB__jaaql, ROLE__jaaql, jaaql_db_password)
            conn = self.jaaql_lookup_connection.get_conn()
            self.jaaql_lookup_connection.execute_script_file(conn, join(get_jaaql_root(), DIR__scripts,
                                                                        "04.install_jaaql_data_structures.generated.sql"))
            self.jaaql_lookup_connection.execute_script_file(conn, join(get_jaaql_root(), DIR__scripts, "05.install_jaaql.exceptions.sql"))
            self.jaaql_lookup_connection.execute_script_file(conn, join(get_jaaql_root(), DIR__scripts, "06.install_jaaql.handwritten.sql"))
            self.jaaql_lookup_connection.commit(conn)
            self.jaaql_lookup_connection.put_conn(conn)

            self.create_account_with_potential_password(self.jaaql_lookup_connection, USERNAME__jaaql, ROLE__jaaql, jaaql_password,
                                                        already_exists=True)
            self.create_account_with_potential_password(self.jaaql_lookup_connection, USERNAME__super_db, ROLE__postgres, super_db_password,
                                                        already_exists=True)
            self.create_account_with_potential_password(self.jaaql_lookup_connection, USERNAME__anonymous, password=PASSWORD__anonymous,
                                                        is_the_anonymous_user=True)

            super_conn_str = PROTOCOL__postgres + username + ":" + db_password + "@" + address + ":" + str(port) + "/"
            self.vault.insert_obj(VAULT_KEY__super_db_credentials, super_conn_str)
            jaaql_lookup_str = PROTOCOL__postgres + ROLE__jaaql + ":" + jaaql_db_password + "@" + address + ":" + str(port) + "/" + DB__jaaql
            self.vault.insert_obj(VAULT_KEY__jaaql_lookup_connection, jaaql_lookup_str)

            if self.is_container and do_reboot:
                self.jaaql_lookup_connection.close()
                print("Rebooting to allow JAAQL config to be shared among workers")
                threading.Thread(target=self.exit_jaaql).start()
            else:
                self.has_installed = True
        else:
            raise HttpStatusException(ERR__already_installed)

    def verification_thread(self, the_queue: Queue):
        print("Starting auth verification thread")
        while True:
            auth_token, ip_address, complete = the_queue.get()
            try:
                self.verify_auth_token(auth_token, ip_address)
                complete.put((True, None, None))
            except HttpStatusException as ex:
                complete.put((False, ex.message, ex.response_code))
            except Exception as ex:
                complete.put((False, str(ex), 500))

    def verify_auth_token_threaded(self, auth_token: str, ip_address: str, complete: Queue):
        try:
            if JAAQLModel.VERIFICATION_QUEUE is None:
                JAAQLModel.VERIFICATION_QUEUE = Queue()
                threading.Thread(target=self.verification_thread, args=[JAAQLModel.VERIFICATION_QUEUE], daemon=True).start()
            JAAQLModel.VERIFICATION_QUEUE.put((auth_token, ip_address, complete))
            payload = json.loads(base64url_decode(auth_token.split(".")[1].encode("UTF-8")).decode())
            return payload[KEY__account_id], payload[KEY__username], payload[KEY__ip_address], payload[KEY__is_the_anonymous_user]
        except Exception:
            raise HttpStatusException(ERR__invalid_token, HTTPStatus.UNAUTHORIZED)

    def verify_auth_token(self, auth_token: str, ip_address: str):
        decoded = crypt_utils.jwt_decode(self.vault.get_obj(VAULT_KEY__jwt_crypt_key), auth_token, JWT_PURPOSE__oauth)
        if not decoded or decoded[KEY__ip_address] != ip_address:
            raise HttpStatusException(ERR__invalid_token, HTTPStatus.UNAUTHORIZED)

        validate_is_most_recent_password(self.jaaql_lookup_connection, decoded[KEY__account_id], decoded[KEY__password],
                                         singleton_message=ERR__invalid_token, singleton_code=HTTPStatus.UNAUTHORIZED)

        return decoded[KEY__account_id], decoded[KEY__username], decoded[KEY__ip_address], decoded[KEY__is_the_anonymous_user]

    def refresh_auth_token(self, auth_token: str, ip_address: str):
        decoded = crypt_utils.jwt_decode(self.vault.get_obj(VAULT_KEY__jwt_crypt_key), auth_token, JWT_PURPOSE__oauth, allow_expired=True)
        if not decoded:
            raise HttpStatusException(ERR__invalid_token, HTTPStatus.UNAUTHORIZED)

        if datetime.fromisoformat(decoded[KEY__created]) + timedelta(milliseconds=self.refresh_expiry_ms) < datetime.now():
            raise HttpStatusException(ERR__refresh_expired, HTTPStatus.UNAUTHORIZED)

        return self.get_auth_token(decoded[KEY__username], ip_address)

    def get_auth_token(self, username: str, ip_address: str, password: str = None, response: JAAQLResponse = None):
        incorrect_credentials = False
        account = None
        encrypted_salted_ip_address = None

        try:
            account = fetch_most_recent_password_from_username(self.jaaql_lookup_connection, self.get_db_crypt_key(),
                                                               self.get_vault_repeatable_salt(), username, singleton_code=HTTPStatus.UNAUTHORIZED)
        except HttpStatusException as exc:
            if exc.response_code == HTTPStatus.UNAUTHORIZED:
                incorrect_credentials = True
            else:
                raise exc

        if not incorrect_credentials:
            salt_user = self.get_repeatable_salt(account[KG__account__id])

            encrypted_salted_ip_address = jaaql__encrypt(ip_address, self.get_db_crypt_key(), salt_user)  # An optimisation, it is used later twice

            if password is not None:
                incorrect_credentials = crypt_utils.verify_password_hash(account[KG__account_password__hash], password,
                                                                         salt=self.get_repeatable_salt(account[KG__account__id]))
            else:
                incorrect_credentials = not exists_matching_validated_ip_address(self.jaaql_lookup_connection, encrypted_salted_ip_address)

        if incorrect_credentials:
            raise HttpStatusException(ERR__incorrect_credentials, HTTPStatus.UNAUTHORIZED)

        address = execute_supplied_statement_singleton(self.jaaql_lookup_connection,
                                                       QUERY___add_or_update_validated_ip_address,
                                                       {KG__validated_ip_address__account: account[KG__account__id],
                                                        KG__validated_ip_address__encrypted_salted_ip_address: encrypted_salted_ip_address},
                                                       as_objects=True)[KG__validated_ip_address__uuid]

        jwt_data = {
            KEY__account_id: str(account[KG__account__id]),
            KEY__password: str(account[KG__account_password__uuid]),
            KEY__username: username,
            KEY__ip_address: ip_address,
            KEY__ip_id: str(address),
            KEY__created: datetime.now().isoformat(),
            KEY__is_the_anonymous_user: username == USERNAME__anonymous
        }

        if response is not None:
            response.account_id = str(account[KG__account__id]),
            response.ip_id = str(address)

        return crypt_utils.jwt_encode(self.vault.get_obj(VAULT_KEY__jwt_crypt_key), jwt_data, JWT_PURPOSE__oauth, expiry_ms=self.token_expiry_ms)

    def create_interface_for_db(self, user_id: str, database: str, sub_role: str = None):
        jaaql_uri = self.vault.get_obj(VAULT_KEY__super_db_credentials)
        address, port, _, username, password = DBInterface.fracture_uri(jaaql_uri)
        return create_interface(self.config, address, port, database, username, password=password, role=user_id, sub_role=sub_role)

    def attach_dispatcher_credentials(self, connection: DBInterface, inputs: dict):
        email_dispatcher__update(connection, self.get_db_crypt_key(), **inputs)

    def send_email(self, application: str, template: str, application_artifacts_source: str, application_base_url: str, account_id: str,
                   parameters: dict = None, parameter_id: str = None, none_sanitized_parameters: dict = None):
        if none_sanitized_parameters is None:
            none_sanitized_parameters = {}

        account = account__select(self.jaaql_lookup_connection, self.get_db_crypt_key(), account_id)
        template = email_template__select(self.jaaql_lookup_connection, application, template)

        if parameters is not None and len(parameters) == 0:
            parameters = None

        if template[KG__email_template__validation_schema] is None and parameters is not None:
            raise HttpStatusException(ERR__unexpected_parameters)
        if template[KG__email_template__validation_schema] is not None and parameters is None:
            raise HttpStatusException(ERR__expected_parameters)

        if parameters is not None:
            ins_query = "INSERT INTO %s (%s) VALUES (%s) RETURNING id"

            if parameter_id is not None:
                parameters[KEY__id] = parameter_id

            for col, _ in parameters.items():
                if not re.match(REGEX__object_name, col):
                    raise HttpStatusException(ERR__invalid_object_name % col)

            cols = ", ".join(['"' + key + '"' for key in parameters.keys()])
            vals = ", ".join([':' + key for key in parameters.keys()])

            ins_query = ins_query % (template[KG__email_template__data_validation_table], cols, vals)
            self.submit({
                KEY__application: application,
                KEY__schema: template[KG__email_template__validation_schema],
                KEY_query: ins_query,
                KEY_parameters: parameters,
                KEY_assert: ASSERT_one
            }, account_id)
            parameter_id = execute_supplied_statement_singleton(self.jaaql_lookup_connection, ins_query, parameters, as_objects=True)[KEY__id]

            sel_table = \
                template[KG__email_template__data_validation_table] if template[KG__email_template__data_validation_view] is None \
                else template[KG__email_template__data_validation_view]

            sel_query = "SELECT * FROM %s WHERE id = :id" % sel_table
            self.submit({
                KEY__application: application,
                KEY__schema: template[KG__email_template__validation_schema],
                KEY_query: sel_query,
                KEY_parameters: parameters,
                KEY_assert: ASSERT_one
            }, account_id)
            parameters = execute_supplied_statement_singleton(self.jaaql_lookup_connection, sel_query, {KEY__id: parameter_id}, as_objects=True)
        else:
            parameters = {}

        none_sanitized_parameters[EMAIL_PARAM__app_url] = application_base_url
        none_sanitized_parameters[EMAIL_PARAM__email_address] = account[KG__account__username]
        parameters = {**parameters, **none_sanitized_parameters}

        self.email_manager.construct_and_send_email(application_artifacts_source, template[KG__email_template__dispatcher], template,
                                                    account[KG__account__username], parameters)

    def gen_security_event_unlock_code(self, codeset: str, length: int):
        return "".join([codeset[random.randint(0, len(codeset) - 1)] for _ in range(length)])

    def check_security_event_key_and_security_event_is_unlocked(self, inputs: dict, returning: bool = False):
        evt = check_security_event_unlock(
            self.jaaql_lookup_connection, inputs[KG__security_event__event_lock], inputs[KG__security_event__unlock_code],
            inputs[KG__security_event__unlock_key], singleton_message=ERR__invalid_lock
        )

        if evt[KG__security_event__wrong_key_attempt_count] >= CODE__max_attempts:
            raise HttpStatusException("Too many invalid attempts to unlock the code")

        timezone = evt[KG__security_event__creation_timestamp].tzinfo
        add_seconds = timedelta(seconds=evt[KG__application__unlock_code_validity_period])
        print(timezone)
        print(evt[KG__security_event__creation_timestamp])
        print(add_seconds)
        print(evt[KG__security_event__creation_timestamp] + add_seconds)
        print(datetime.now(timezone))
        if evt[KG__security_event__creation_timestamp] + add_seconds < datetime.now(timezone):
            raise HttpStatusException(ERR__unlock_code_expired)

        if not evt[KEY__key_fits]:
            raise HttpStatusException(ERR__incorrect_lock_code)

        if returning:
            return evt
        else:
            template = email_template__select(self.jaaql_lookup_connection, evt[KG__security_event__application],
                                              evt[KG__security_event__email_template])
            return template[KG__email_template__type]

    def mark_security_event_unlocked(self, sec_evt: dict):
        security_event__update(self.jaaql_lookup_connection, sec_evt[KG__security_event__application], sec_evt[KG__security_event__event_lock],
                               unlock_timestamp=datetime.now())

    def finish_security_event(self, inputs: dict, ip_address: str):
        sec_evt = self.check_security_event_key_and_security_event_is_unlocked({
            KG__security_event__event_lock: inputs[KG__security_event__event_lock],
            KG__security_event__unlock_code: inputs[KG__security_event__unlock_code],
            KG__security_event__unlock_key: inputs[KG__security_event__unlock_key]
        }, returning=True)

        template = email_template__select(self.jaaql_lookup_connection, sec_evt[KG__security_event__application],
                                          sec_evt[KG__security_event__email_template])

        parameters = None

        if template[KG__email_template__data_validation_table]:
            app = application__select(self.jaaql_lookup_connection, sec_evt[KG__security_event__application])  # TODO try select parameters
            parameters = {}
            # TODO possibly load parameters

        if template[KG__email_template__type] in [EMAIL_TYPE__signup, EMAIL_TYPE__reset_password, EMAIL_TYPE__unregistered_password_reset]:
            self.add_account_password(sec_evt[KG__security_event__account], inputs[KEY__password])

        account = account__select(self.jaaql_lookup_connection, self.get_db_crypt_key(), sec_evt[KG__security_event__account])
        auth_token = self.get_auth_token(account[KG__account__username], ip_address, inputs[KEY__password])

        self.mark_security_event_unlocked(sec_evt)

        return {
            KEY__oauth_token: auth_token,
            KEY__parameters: parameters
        }

    def reset_password(self, inputs: dict):
        app = application__select(self.jaaql_lookup_connection, inputs[KG__security_event__application])
        if inputs[KEY__reset_password_template] is None:
            inputs[KEY__reset_password_template] = app[KG__application__default_r_et]
        if inputs[KEY__unregistered_user_reset_password_template] is None:
            inputs[KEY__unregistered_user_reset_password_template] = app[KG__application__default_u_et]

        if inputs[KEY__reset_password_template] is None:
            raise HttpStatusException("Missing reset password template for application. Either supply one in the reset call or set a default")

        if inputs[KEY__unregistered_user_reset_password_template] is None:
            raise HttpStatusException("Missing unregistered user template for application. Either supply one in the reset call or set a default")

        reset_template = email_template__select(self.jaaql_lookup_connection, inputs[KG__security_event__application],
                                                inputs[KEY__reset_password_template])
        if reset_template[KG__email_template__type] != EMAIL_TYPE__reset_password:
            raise HttpStatusException(ERR__template_not_reset)

        unregistered_template = email_template__select(self.jaaql_lookup_connection, inputs[KG__security_event__application],
                                                       inputs[KEY__unregistered_user_reset_password_template])
        if unregistered_template[KG__email_template__type] != EMAIL_TYPE__unregistered_password_reset:
            raise HttpStatusException(ERR__template_not_unregistered)

        account_existed = False
        try:
            account_id = self.create_account_with_potential_password(self.jaaql_lookup_connection, inputs[KEY__username])
        except HttpStatusException as hs:
            if not hs.message.startswith(SQL__err_duplicate):
                raise hs  # Unrelated exception, raise it

            account = fetch_account_from_username(self.jaaql_lookup_connection, self.get_db_crypt_key(), self.get_vault_repeatable_salt(),
                                                  inputs[KEY__username])
            account_id = account[KG__account__id]
            if account[KG__account__most_recent_password] is not None:
                account_existed = True

        count = execute_supplied_statement_singleton(self.jaaql_lookup_connection,
                                                     QUERY__count_security_events_of_type_in_24hr_window,
                                                     {
                                                         KEY__type_one: EMAIL_TYPE__reset_password,
                                                         KEY__type_two: EMAIL_TYPE__unregistered_password_reset,
                                                         KG__security_event__account: account_id
                                                     }, as_objects=True)[ATTR__count]

        if count >= RESEND__reset_max:
            raise HttpStatusException(ERR__too_many_reset_requests, HTTPStatus.TOO_MANY_REQUESTS)

        template = None
        if account_existed and inputs[KEY__reset_password_template] is not None:
            template = inputs[KEY__reset_password_template]
        elif not account_existed and inputs[KEY__unregistered_user_reset_password_template] is not None:
            template = inputs[KEY__unregistered_user_reset_password_template]

        unlock_code = self.gen_security_event_unlock_code(CODE__letters, CODE__invite_length)
        reg_env_ins = security_event__insert(self.jaaql_lookup_connection, inputs[KG__security_event__application], template, account_id,
                                             unlock_code)

        self.send_email(inputs[KG__security_event__application], template, app[KG__application__artifacts_source],
                        app[KG__application__base_url], account_id, inputs[KEY__parameters],
                        parameter_id=reg_env_ins[KG__security_event__event_lock], none_sanitized_parameters={
                EMAIL_PARAM__unlock_key: reg_env_ins[KG__security_event__unlock_key],
                EMAIL_PARAM__unlock_code: unlock_code
            })

        return {
            KG__security_event__event_lock: reg_env_ins[KG__security_event__event_lock]
        }

    def sign_up(self, inputs: dict):
        app = application__select(self.jaaql_lookup_connection, inputs[KG__security_event__application])
        if inputs[KEY__sign_up_template] is None:
            inputs[KEY__sign_up_template] = app[KG__application__default_s_et]
        if inputs[KEY__already_signed_up_template] is None:
            inputs[KEY__already_signed_up_template] = app[KG__application__default_a_et]

        if inputs[KEY__sign_up_template] is None:
            raise HttpStatusException("Missing sign up template for application. Either supply one in the sign up call or set a default")

        if inputs[KEY__already_signed_up_template] is None:
            raise HttpStatusException("Missing already signed up template for application. Either supply one in the sign up call or set a default")

        sign_up_template = email_template__select(self.jaaql_lookup_connection, inputs[KG__security_event__application],
                                                  inputs[KEY__sign_up_template])
        if sign_up_template[KG__email_template__type] != EMAIL_TYPE__signup:
            raise HttpStatusException(ERR__template_not_signup)

        already_signed_up_template = email_template__select(self.jaaql_lookup_connection, inputs[KG__security_event__application],
                                                            inputs[KEY__already_signed_up_template])
        if already_signed_up_template[KG__email_template__type] != EMAIL_TYPE__already_signed_up:
            raise HttpStatusException(ERR__template_not_already)

        account_existed = False
        try:
            account_id = self.create_account_with_potential_password(self.jaaql_lookup_connection, inputs[KEY__username])
        except HttpStatusException as hs:
            if not hs.message.startswith(SQL__err_duplicate):
                raise hs  # Unrelated exception, raise it

            account = fetch_account_from_username(self.jaaql_lookup_connection, self.get_db_crypt_key(), self.get_vault_repeatable_salt(),
                                                  inputs[KEY__username])
            account_id = account[KG__account__id]
            if account[KG__account__most_recent_password] is not None:
                account_existed = True

        count = execute_supplied_statement_singleton(self.jaaql_lookup_connection,
                                                     QUERY__count_security_events_of_type_in_24hr_window,
                                                     {
                                                         KEY__type_one: EMAIL_TYPE__signup,
                                                         KEY__type_two: EMAIL_TYPE__already_signed_up,
                                                         KG__security_event__account: account_id
                                                     }, as_objects=True)[ATTR__count]

        if count >= RESEND__invite_max:
            raise HttpStatusException(ERR__too_many_signup_attempts, HTTPStatus.TOO_MANY_REQUESTS)

        template = None
        if account_existed and inputs[KEY__already_signed_up_template] is not None:
            template = inputs[KEY__already_signed_up_template]
        elif not account_existed and inputs[KEY__sign_up_template] is not None:
            template = inputs[KEY__sign_up_template]

        unlock_code = self.gen_security_event_unlock_code(CODE__letters, CODE__invite_length)
        reg_env_ins = security_event__insert(self.jaaql_lookup_connection, inputs[KG__security_event__application], template, account_id,
                                             unlock_code)

        self.send_email(inputs[KG__security_event__application], template, app[KG__application__artifacts_source],
                        app[KG__application__base_url], account_id, inputs[KEY__parameters],
                        parameter_id=reg_env_ins[KG__security_event__event_lock], none_sanitized_parameters={
                EMAIL_PARAM__unlock_key: reg_env_ins[KG__security_event__unlock_key],
                EMAIL_PARAM__unlock_code: unlock_code
            })

        return {
            KG__security_event__event_lock: reg_env_ins[KG__security_event__event_lock]
        }

    def submit(self, inputs: dict, account_id: str, verification_hook: Queue = None):
        if not isinstance(inputs, dict):
            raise HttpStatusException("Expected object or string input")

        if KEY__application in inputs:
            schemas = execute_supplied_statement(self.jaaql_lookup_connection, QUERY__fetch_application_schemas, {
                KG__application_schema__application: inputs[KEY__application]
            }, as_objects=True)
            if len(schemas) == 0:
                raise HttpStatusException("Application has no schemas!")
            if not schemas[0][KG__application__is_live]:
                raise HttpStatusException("Application is currently being deployed. Please wait a few minutes until deployment is complete")
            schemas = {itm[KG__application_schema__name]: itm for itm in schemas}

            found_db = None
            if KEY__schema in inputs:
                found_db = schemas[inputs[KEY__schema]][KEY__database]
                inputs.pop(KEY__schema)
            else:
                if len(schemas) == 1:
                    found_db = schemas[list(schemas.keys())[0]][KEY__database]
                else:
                    found_dbs = [val[KEY__database] for _, val in schemas.items() if val[KEY__is_default]]
                    if len(found_dbs) == 1:
                        found_db = found_dbs[0]

            if not found_db:
                raise HttpStatusException(ERR__schema_invalid)

            inputs[KEY__database] = found_db

        if KEY__database not in inputs:
            inputs[KEY__database] = DB__jaaql

        sub_role = inputs.pop(KEY__role) if KEY__role in inputs else None
        required_db = self.create_interface_for_db(account_id, inputs[KEY__database], sub_role)

        return InterpretJAAQL(required_db).transform(inputs, skip_commit=inputs.get(KEY__read_only), wait_hook=verification_hook,
                                                     encryption_key=self.get_db_crypt_key(),
                                                     canned_query_service=self.cached_canned_query_service)



from jaaql.openapi.swagger_documentation import *
from jaaql.constants import *
from jaaql.mvc.generated_queries import *

TITLE = "JAAQL Internal API"
DESCRIPTION = "Collection of methods in the JAAQL internal API"
FILENAME = "jaaql_internal_api"

ARG_RES__install_key = SwaggerArgumentResponse(
    name=KEY__install_key,
    description="Single use/purpose key. Find in the logs, line starting with 'INSTALL KEY:' at startup",
    arg_type=str,
    example=["aefc1b08-a573-466f-bdd0-706ae281cc99", "ec63aa9a-b189-419f-8e0b-7fcc4ff8c857"],
    required=True
)

DOCUMENTATION__install = SwaggerDocumentation(
    tags="Installation",
    security=False,  # This method is not secured as the system is not setup yet. It uses a OTP token system via logs
    methods=SwaggerMethod(
        name="Install JAAQL",
        description="Installs JAAQL to the configured database. Please allow a minute after running for the server to "
                    "refresh and reload. A connection string is required for local installation and is not allowed when running "
                    "inside a docker container as the docker container comes with a postgres database and JAAQL will use that",
        method=REST__POST,
        body=[
            SwaggerArgumentResponse(
                name="db_connection_string",
                description="Database connection string for the superuser",
                arg_type=str,
                example=["postgresql://postgres:123456@localhost:5432/",
                         "postgresql://postgres:pa55word@localhost:5432/"],
                required=True,
                local_only=True
            ),
            ARG_RES__install_key,
            SwaggerArgumentResponse(
                name=KEY__jaaql_password,
                description="The password for the jaaql user, can create applications, configurations",
                example=["pa55word"],
                strip=True,
                arg_type=str
            ),
            SwaggerArgumentResponse(
                name=KEY__super_db_password,
                description="Gives access to the database super user",
                example=["passw0rd"],
                required=True,
                strip=True,
                arg_type=str
            ),
            SwaggerArgumentResponse(
                name=KEY__allow_uninstall,
                description="Allows uninstall. DO NOT USE ON PROD, used to aid with testing. Defaults to false if not supplied",
                example=[False, True],
                required=False,
                condition="If you want to specify if uninstallation allowed",
                arg_type=bool
            )
        ],
        response=[
            SwaggerFlatResponse(),
            SwaggerFlatResponse(
                description=ERR__already_installed,
                code=HTTPStatus.UNPROCESSABLE_ENTITY,
                body=ERR__already_installed
            )
        ]
    )
)

DOCUMENTATION__is_installed = SwaggerDocumentation(
    tags="Installation",
    security=False,
    methods=SwaggerMethod(
        name="Is installed",
        description="Returns 200 OK if the service has installed otherwise a 422",
        method=REST__GET,
        response=[
            SwaggerFlatResponse(),
            SwaggerFlatResponse(
                description=ERR__not_yet_installed,
                code=HTTPStatus.UNPROCESSABLE_ENTITY,
                body=ERR__not_yet_installed
            )
        ]
    )
)

DOCUMENTATION__is_alive = SwaggerDocumentation(
    security=False,
    tags="Is Alive",
    methods=SwaggerMethod(
        name="Check is alive",
        description="An endpoint that can be called by any service to see if the service is alive",
        method=REST__GET
    )
)

DOCUMENTATION__clean = SwaggerDocumentation(
    tags="Installation",
    methods=SwaggerMethod(
        name="Clean database",
        description="Will clear the database and leave jaaql installed",
        method=REST__POST
    )
)

# Not unused. Used to generate html files
from jaaql.documentation.documentation_shared import DOCUMENTATION__oauth_token, DOCUMENTATION__oauth_refresh

DOCUMENTATION__dispatchers = SwaggerDocumentation(
    tags="Dispatchers",
    methods=SwaggerMethod(
        name="Attach credentials to dispatcher",
        description="Given a prime key for a dispatcher, will encrypt credentials and attach at the database level",
        method=REST__POST,
        body=[
            SwaggerArgumentResponse(
                name=KG__email_dispatcher__application,
                description="The application of the dispatcher",
                arg_type=str,
                example=["out-and-about"]
            ),
            SwaggerArgumentResponse(
                name=KG__email_dispatcher__name,
                description="The name of the email dispatcher",
                arg_type=str,
                example="default"
            ),
            SwaggerArgumentResponse(
                name=KG__email_dispatcher__url,
                description="The url of the mail server against which the dispatcher will authenticate",
                arg_type=str,
                example="mail-server.nl"
            ),
            SwaggerArgumentResponse(
                name=KG__email_dispatcher__port,
                description="The port of the email server which you can connect to using SMTP STARTTLS",
                arg_type=int,
                example=587
            ),
            SwaggerArgumentResponse(
                name=KG__email_dispatcher__username,
                description="The username with which the dispatcher will authenticate",
                arg_type=str,
                example="address@domain.com"
            ),
            SwaggerArgumentResponse(
                name=KG__email_dispatcher__password,
                description="The password with which the dispatcher will authenticate",
                arg_type=str,
                example="pa55word"
            )
        ]
    )
)



from jaaql.mvc.model import JAAQLModel
from jaaql.mvc.base_controller import BaseJAAQLController
from jaaql.documentation.documentation_internal import *
from jaaql.documentation.documentation_public import *
from jaaql.documentation.documentation_shared import *
from jaaql.mvc.response import JAAQLResponse
from jaaql.db.db_interface import DBInterface
import queue


class JAAQLController(BaseJAAQLController):

    def __init__(self, model: JAAQLModel, is_prod: bool, base_url: str, do_profiling: bool = False):
        super().__init__(model, is_prod, base_url, do_profiling)

    def create_app(self):

        @self.cors_route('/oauth/token', DOCUMENTATION__oauth_token)
        def fetch_oauth_token(http_inputs: dict, ip_address: str, response: JAAQLResponse):
            return self.model.get_auth_token(**http_inputs, ip_address=ip_address, response=response)

        @self.cors_route(ENDPOINT__refresh, DOCUMENTATION__oauth_refresh)
        def refresh_oauth_token(auth_token: str, ip_address: str):
            return self.model.refresh_auth_token(auth_token, ip_address)

        @self.cors_route(ENDPOINT__install, DOCUMENTATION__install)
        def install(http_inputs: dict):
            return self.model.install(**http_inputs)

        @self.cors_route('/internal/is_installed', DOCUMENTATION__is_installed)
        def is_installed(response: JAAQLResponse):
            return self.model.is_installed(response)

        @self.cors_route(ENDPOINT__is_alive, DOCUMENTATION__is_alive)
        def is_alive():
            self.model.is_alive()

        @self.cors_route('/accounts', DOCUMENTATION__create_account)
        def accounts(connection: DBInterface, http_inputs: dict):
            self.model.create_account_with_potential_password(connection, **http_inputs)

        @self.cors_route('/account/password', DOCUMENTATION__password)
        def password(account_id: str, username: str, ip_address: str, http_inputs: dict):
            return self.model.add_my_account_password(account_id, username, ip_address, **http_inputs)

        @self.cors_route('/submit', DOCUMENTATION__submit)
        def submit(http_inputs: dict, account_id: str, verification_hook: queue.Queue):
            return self.model.submit(http_inputs, account_id, verification_hook=verification_hook)

        @self.cors_route('/internal/clean', DOCUMENTATION__clean)
        def clean(connection: DBInterface):
            self.model.clean(connection)

        @self.cors_route('/internal/dispatchers', DOCUMENTATION__dispatchers)
        def dispatchers(connection: DBInterface, http_inputs: dict):
            self.model.attach_dispatcher_credentials(connection, http_inputs)

        @self.cors_route('/sign-up', DOCUMENTATION__sign_up)
        def sign_up(http_inputs: dict):
            return self.model.sign_up(http_inputs)

        @self.cors_route('/account/reset-password', DOCUMENTATION__reset_password)
        def reset_password(http_inputs: dict):
            return self.model.reset_password(http_inputs)

        @self.cors_route('/security-event', DOCUMENTATION__security_event)
        def security_event(http_inputs: dict, ip_address: str):
            if self.is_post():
                return self.model.check_security_event_key_and_security_event_is_unlocked(http_inputs)
            else:
                return self.model.finish_security_event(http_inputs, ip_address)


import re

from typing import Callable
from jaaql.email.email_manager_service import Email, TYPE__email_attachments
import requests
from jaaql.exceptions.http_status_exception import HttpStatusException
from jaaql.constants import *
from urllib.parse import quote
from jaaql.utilities.utils_no_project_imports import load_artifact
from typing import Optional
from jaaql.mvc.handmade_queries import *

REGEX__email_parameter = r'({{)([a-zA-Z0-9_\-]+)(}})'
REGEX__email_uri_encoded_parameter = r'(\[\[)([a-zA-Z0-9_\-]+)(\]\])'
REPLACE__str = "{{%s}}"
REPLACE__uri_encoded_str = "[[%s]]"
ERR__missing_parameter = "Missing parameter from template '%s'"
ERR__unexpected_parameter_in_template = "Unexpected parameter in template '%s'"

EMAIL_HTML__start = "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\r\n<html xmlns=\"http://www.w3.org/1999/xhtml\" lang=\"en\">\r\n<body>\r\n"
EMAIL_HTML__end = "\r\n</body>\r\n</html>"

SUBJECT_MARKER = "Subject: "


class EmailManager:

    def __init__(self, is_container: bool):
        self.is_container = is_container

    def send_email(self, email: Email):
        requests.post("http://127.0.0.1:" + str(PORT__ems) + ENDPOINT__send_email, json=email.repr_json())

    @staticmethod
    def uri_encode_replace(val):
        return quote(str(val))

    def construct_and_send_email(self, application_artifacts_source: Optional[str], dispatcher: str, template: dict, to_email: str,
                                 parameters: Optional[dict], attachments=None, attachment_access_token: str = None):
        loaded_template = load_artifact(self.is_container, application_artifacts_source, template[KG__email_template__content_url])

        loaded_template = self.perform_replacements(loaded_template, REPLACE__str, str, REGEX__email_parameter, parameters)
        loaded_template = self.perform_replacements(loaded_template, REPLACE__uri_encoded_str, EmailManager.uri_encode_replace,
                                                    REGEX__email_uri_encoded_parameter, parameters)

        first_line = loaded_template.split("\n")[0].strip()
        if SUBJECT_MARKER not in first_line:
            raise HttpStatusException("Email template does not have a subject on the first line denoted by 'Subject: '")
        subject = first_line.split(SUBJECT_MARKER)[1].strip()
        body = "\n".join(loaded_template.split("\n")[1:]).strip()

        if template[KG__email_template__content_url].lower().endswith(".htmlbody"):
            body = EMAIL_HTML__start + body + EMAIL_HTML__end

        if attachments is not None:
            attachment_list = attachments
            if not isinstance(attachment_list, list):
                attachment_list = [attachment_list]
            for attachment in attachment_list:
                attachment.format_attachment_url(application_artifacts_source, self.is_container)

        self.send_email(Email(template[KEY__application], template[KG__email_template__name], dispatcher, to_email, subject, body,
                              attachments=attachments, attachment_access_token=attachment_access_token))

    def perform_replacements(self, html_template: str, replace_str: str, replace_func: Callable, replace_regex: str, args: dict = None):
        if args is None:
            args = {}

        for key, val in args.items():
            html_template = html_template.replace(replace_str % key.upper(), replace_func(val))

        matched = re.findall(replace_regex, html_template)
        if len(matched) != 0:
            raise HttpStatusException(ERR__missing_parameter % matched[0][1])

        return html_template



from jaaql.openapi.swagger_documentation import SwaggerDocumentation, SwaggerMethod, SwaggerArgumentResponse,\
    SwaggerFlatResponse, REST__POST, ARG_RESP__allow_all, RES__allow_all
from jaaql.constants import *
from typing import Union


OUTPUT = False

ENDPOINT__refresh = "/oauth/refresh"


def set_nullable(arg_res: Union[SwaggerArgumentResponse, list], condition: Union[str, list] = None,
                 new_name: Union[str, list] = None):
    if isinstance(arg_res, list):
        if new_name is None:
            new_name = [None] * len(arg_res)
        return [set_nullable(cur, condition) for cur, condition, new_name in zip(arg_res, condition, new_name)]
    else:
        return SwaggerArgumentResponse(
            arg_res.name if new_name is None else new_name,
            arg_res.description,
            arg_res.arg_type,
            arg_res.example,
            False,
            condition=condition
        )


def rename_arg(arg_res: SwaggerArgumentResponse, new_name: str = None, new_description: str = None, new_examples=None):
    return SwaggerArgumentResponse(
        arg_res.name if new_name is None else new_name,
        arg_res.description if new_description is None else new_description,
        arg_res.arg_type,
        arg_res.example if new_examples is None else new_examples,
        arg_res.required,
        arg_res.condition
    )


RES__oauth_token = SwaggerFlatResponse(
    description="A temporary JWT token that can be used to authenticate with the server",
    body=EXAMPLE__jwt
)

ARG_RES__username = SwaggerArgumentResponse(
    name=KEY__username,
    description="JAAQL login username",
    arg_type=str,
    lower=True,
    strip=True,
    example=["jaaql", "aaron@jaaql.com"]
)

EXAMPLE__password = ["pa55word", "p@ssword"]

ARG_RES__password = SwaggerArgumentResponse(
    name=KEY__password,
    description="JAAQL login password",
    arg_type=str,
    strip=True,
    example=EXAMPLE__password,
    required=True
)

DOCUMENTATION__oauth_token = SwaggerDocumentation(
    tags="OAuth",
    security=False,  # This _is_ the security method, therefore it is not expecting a jwt token
    methods=SwaggerMethod(
        name="OAuth Fetch Token",
        description="Authenticate with the server. Send username and password and server will respond with 200 and a "
                    "token which can be used to access the service. The server may also respond with a 202 and a "
                    "token, this indicates that an mfa key is expected. Send the token back to the service along with "
                    "an MFA key and you will returned the aforementioned 200 response",
        method=REST__POST,
        body=[
            ARG_RES__username,
            ARG_RES__password
        ],
        response=RES__oauth_token
    )
)

DOCUMENTATION__oauth_refresh = SwaggerDocumentation(
    tags="OAuth",
    security=True,
    methods=SwaggerMethod(
        name="OAuth Refresh Token",
        description="Refresh your token",
        method=REST__POST,
        response=RES__oauth_token
    )
)

DOCUMENTATION__submit = SwaggerDocumentation(
    tags="Queries",
    methods=SwaggerMethod(
        name="Execute JAAQL query",
        description="Executes a JAAQL query which is either a single SQL query or a list of queries. Returns results",
        method=REST__POST,
        arguments=ARG_RESP__allow_all,
        response=RES__allow_all,
        parallel_verification=True
    )
)



from jaaql.openapi.swagger_documentation import *
from jaaql.constants import *
from jaaql.mvc.generated_queries import *
from jaaql.documentation.documentation_shared import ARG_RES__username, set_nullable, ARG_RES__password, RES__oauth_token, rename_arg

TITLE = "JAAQL API"
DESCRIPTION = "Collection of methods in the JAAQL API"
FILENAME = "jaaql_api"

# Not unused. Used to generate html files
from jaaql.documentation.documentation_shared import DOCUMENTATION__oauth_token, DOCUMENTATION__oauth_refresh

DOCUMENTATION__create_account = SwaggerDocumentation(
    tags="Admin",
    methods=SwaggerMethod(
        method=REST__POST,
        name="Create account",
        description="Will create an account, if you have privileges to do so",
        body=[
            ARG_RES__username,
            set_nullable(ARG_RES__password, "Whether the user is given a password"),
            SwaggerArgumentResponse(
                name=KEY__attach_as,
                description="Whether the user will attach as a role",
                arg_type=str,
                required=False,
                condition="Defaults to false",
                example="my-role"
            )
        ]
    )
)

DOCUMENTATION__password = SwaggerDocumentation(
    tags="Account",
    methods=SwaggerMethod(
        method=REST__POST,
        name="Change password",
        description="Will change the logged in users password to the new one",
        arguments=[
            rename_arg(ARG_RES__username, KEY__old_password, "The old password for the user"),
            ARG_RES__password
        ],
        response=RES__oauth_token
    )
)

ARG_RES__event_application = SwaggerArgumentResponse(
    name=KG__security_event__application,
    description="The application of the dispatcher",
    arg_type=str,
    example=["out-and-about"]
)

ARG_RES__parameters = SwaggerArgumentResponse(
    name=KEY__parameters,
    description="Nonspecific data which is supplied as an object either for email or signup. Is validated",
    arg_type=ARG_RESP__allow_all,
    required=False,
    condition="Is signup data provided"
)

ARG_RES__event_lock = SwaggerArgumentResponse(
    name=KG__security_event__event_lock,
    description="A key that is used to reference a security event",
    arg_type=str,
    example="0aff271e-bf0a-463b-b234-f558abd70edd"
)

ARG_RES__security_event_unlock_key = SwaggerArgumentResponse(
    name=KG__security_event__unlock_key,
    description="The unlock key",
    required=False,
    condition="Are you using a key to unlock or a code",
    arg_type=str,
    example="c9a941d5-2aea-4b96-967b-32a01e889d03"
)

ARG_RES__security_event_unlock_code = SwaggerArgumentResponse(
    name=KG__security_event__unlock_code,
    description="The unlock code",
    required=False,
    condition="Are you using a key to unlock or a code",
    arg_type=str,
    example="ABC123"
)

KEY__oauth_token = "oauth_token"

ARG_RES__oauth_token = SwaggerArgumentResponse(
    name=KEY__oauth_token,
    description="An auth token",
    arg_type=str,
    example=EXAMPLE__jwt
)

DOCUMENTATION__reset_password = SwaggerDocumentation(
    tags="Reset Password",
    security=False,
    methods=[
        SwaggerMethod(
            name="Request an password reset",
            description="Requests a password reset, sending an email to the user (or if the email template does not contain "
            "the reset code, they will then need to be approved by an admin).",
            method=REST__POST,
            body=[
                ARG_RES__username,
                ARG_RES__parameters,
                ARG_RES__event_application,
                SwaggerArgumentResponse(
                    name=KEY__reset_password_template,
                    description="The template to send if the user is already signed up",
                    required=False,
                    condition="Should an email be sent or the default be used",
                    arg_type=str,
                    example=KG__application__default_r_et
                ),
                SwaggerArgumentResponse(
                    name=KEY__unregistered_user_reset_password_template,
                    description="The template to send if the user hasn't signed up",
                    required=False,
                    condition="Should an email be sent or the default be used",
                    arg_type=str,
                    example=KG__application__default_u_et
                )
            ],
            response=[
                SwaggerResponse(
                    description="Reset password response",
                    response=ARG_RES__event_lock
                ),
                SwaggerFlatResponse(
                    description=ERR__too_many_reset_requests,
                    code=HTTPStatus.TOO_MANY_REQUESTS,
                    body=ERR__too_many_reset_requests
                )
            ]
        )
    ]
)

DOCUMENTATION__sign_up = SwaggerDocumentation(
    tags="Signup",
    security=False,
    methods=[
        SwaggerMethod(
            name="Request an invitation",
            description="Requests an invitation, sending an email to the user (or if the email template does not contain "
            "the unlock code, they will then need to be approved by an admin).",
            method=REST__POST,
            body=[
                ARG_RES__username,
                ARG_RES__parameters,
                ARG_RES__event_application,
                SwaggerArgumentResponse(
                    name=KEY__sign_up_template,
                    description="The template to send if the user hasn't signed up",
                    required=False,
                    condition="Should an email be sent or the default be used",
                    arg_type=str,
                    example=KG__application__default_s_et
                ),
                SwaggerArgumentResponse(
                    name=KEY__already_signed_up_template,
                    description="The template to send if the user is already signed up",
                    required=False,
                    condition="Should an email be sent or the default be used",
                    arg_type=str,
                    example=KG__application__default_a_et
                )
            ],
            response=[
                SwaggerResponse(
                    description="Sign up response",
                    response=ARG_RES__event_lock
                ),
                SwaggerFlatResponse(
                    description=ERR__too_many_signup_attempts,
                    code=HTTPStatus.TOO_MANY_REQUESTS,
                    body=ERR__too_many_signup_attempts
                )
            ]
        )
    ]
)

from jaaql.mvc.handmade_queries import EMAIL_TYPE__signup

DOCUMENTATION__security_event = SwaggerDocumentation(
    tags="Security Event",
    security=False,
    methods=[
        SwaggerMethod(
            name="Check Event and Key",
            description="Checks a security event is unused and that the key is valid",
            method=REST__POST,
            body=[
                ARG_RES__event_lock,
                ARG_RES__security_event_unlock_key,
                ARG_RES__security_event_unlock_code
            ],
            response=SwaggerFlatResponse(
                description="Will receive the type of the description if unlock code is correct",
                body=EMAIL_TYPE__signup
            )
        ),
        SwaggerMethod(
            name="Finish Security Event",
            description="Finishes the security event, supplying a password which can either be a new password or your login password, if using the "
            "already signed up flow",
            method=REST__PUT,
            body=[
                ARG_RES__event_lock,
                ARG_RES__security_event_unlock_key,
                ARG_RES__security_event_unlock_code,
                ARG_RES__password
            ],
            response=SwaggerResponse(
                description="Signup parameters",
                response=[
                    ARG_RES__parameters,
                    ARG_RES__oauth_token
                ]
            )
        )
    ]
)



from datetime import datetime
import os
from jaaql.constants import ENVIRON__install_path, ARTIFACTS_DEFAULT_DIRECTORY
from os.path import join
from jaaql.exceptions.http_status_exception import HttpStatusException
import requests
from typing import Union
import re

ALLOWABLE_FILE_PATH = r'^[a-z0-9_\-\/]+(\.[a-zA-Z0-9]+)?$'


def time_delta_ms(start_time: datetime, end_time: datetime) -> int:
    return int(round((end_time - start_time).total_seconds() * 1000))


def load_artifact(is_container: bool, artifact_base_url: str, app_relative_path: str):
    if app_relative_path is None:
        return None

    if artifact_base_url is None:
        artifact_base_url = ARTIFACTS_DEFAULT_DIRECTORY

    if check_allowable_file_path(app_relative_path):
        raise Exception("Database has been tampered with! Cannot send email")
    if not artifact_base_url.startswith("https://") and not artifact_base_url.startswith("http://"):
        if is_container:
            if check_allowable_file_path(artifact_base_url):
                print(artifact_base_url)
                raise Exception("Illegal artifact source directory")
        if artifact_base_url.startswith("file:///"):
            return open(join(artifact_base_url.split("file:///")[1].replace("%20", " "), app_relative_path), "r").read()
        else:
            base_path = os.environ.get(ENVIRON__install_path)
            if base_path is None:
                base_path = ""
            template_path = join(base_path, "www", artifact_base_url, app_relative_path)
            try:
                return open(template_path.replace("\\", "/"), "r").read()
            except FileNotFoundError:
                raise HttpStatusException("Could not find template at path '%s'. Are you sure the template is accessible to JAAQL?" % template_path)
    else:
        splitter = "" if artifact_base_url.endswith("/") else "/"
        return requests.get(artifact_base_url + splitter + app_relative_path).text


def check_allowable_file_path(uri):
    return re.match(ALLOWABLE_FILE_PATH, uri) is None


def pull_from_dict(self, inputs: dict, keys: Union[list, str, dict]):
    if not isinstance(keys, list) and not isinstance(keys, dict):
        keys = [keys]
    if isinstance(keys, list):
        return {key: inputs[key] for key in keys}
    else:
        return {map_to: inputs[map_from if map_from is not None else map_to] for map_from, map_to in keys.items()}



from jaaql.db.db_interface import DBInterface
from jaaql.db.db_utils import execute_supplied_statement_singleton, execute_supplied_statement
from jaaql.constants import KEY__application, FILE__canned_queries
from jaaql.utilities.utils_no_project_imports import load_artifact
from jaaql.exceptions.http_status_exception import HttpStatusException
from requests.exceptions import RequestException
import json

ERR__invalid_query = "Could not find the canned query that you have specified! Could not find %s"


class CachedCannedQueryService:

    def __init__(self, is_container: bool, connection: DBInterface):
        self.canned_queries = {}
        # self.init_canned_queries(is_container, connection)

    def init_canned_queries(self, is_container: bool, connection: DBInterface):
        pass
        # TODO disabled for now
        # apps = execute_supplied_statement(connection, QUERY__load_application, as_objects=True)
        # for app in apps:
        #     self.refresh_application(is_container, connection, app[KEY__name], app[KEY__artifacts_source])

    def get_canned_query(self, application: str, file: str, pos: int) -> str:
        if application not in self.canned_queries:
            raise HttpStatusException(ERR__invalid_query % "application")

        queries = self.canned_queries[application]

        if file not in queries:
            raise HttpStatusException(ERR__invalid_query % "file")

        if pos not in queries[file]:
            raise HttpStatusException(ERR__invalid_query % "pos")

        return queries[file][pos]

    def refresh_application(self, is_container: bool, connection: DBInterface, application: str, config_resource_url: str = None):
        pass
        # TODO implement in future
        # if config_resource_url is None:
        #     config_resource_url = execute_supplied_statement_singleton(connection, QUERY__load_applications,
        #                                                                {KEY__application: application},
        #                                                                as_objects=True)[KEY__artifact_base_url]
#
        # canned_queries = None
        # try:
        #     canned_queries = load_artifact(is_container, config_resource_url, FILE__canned_queries)
        #     json.loads(canned_queries)
        # except FileNotFoundError:
        #     pass  # No canned queries for app. This is _okay_
        # except RequestException:
        #     pass  # Same as above
#
        # if application not in canned_queries:
        #     self.canned_queries[application] = {}
#
        # self.canned_queries[application] = canned_queries
#
        # TODO broadcast this to other workers
